#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        srv_sim_runner.py
# Purpose:     Manager of the simulation sim_tasks on the server side
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     23-02-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
import threading
import time
from typing import Union, Any, Optional
from pathlib import Path
import zipfile
import logging

from ..sim.sim_runner import SimRunner
from ..editor.base_editor import BaseEditor

_logger = logging.getLogger("spicelib.ServerSimRunner")


def zip_files(raw_filename: Optional[Path], log_filename: Optional[Path]) -> Optional[Path]:
    """Zips the raw and log files into a single zip file.

    :param raw_filename: The path to the raw file.
    :type raw_filename: Optional[pathlib.Path]
    :param log_filename: The path to the log file.
    :type log_filename: Optional[pathlib.Path]
    :return: The path to the zip file or None if no files were provided.
    :rtype: Optional[Path]
    """
    zip_filename = None
    if isinstance(raw_filename, Path) and raw_filename.exists():
        zip_filename = raw_filename.with_suffix('.zip')
    else:
        if isinstance(log_filename, Path) and log_filename.exists():
            zip_filename = log_filename.with_suffix('.zip')
            
    if zip_filename is None:
        _logger.warning("No files are available for zipping. Returning None.")
        return None

    with zipfile.ZipFile(zip_filename, 'w', zipfile.ZIP_DEFLATED) as zip_file:
        if isinstance(raw_filename, Path) and raw_filename.exists():
            zip_file.write(raw_filename)
        if isinstance(log_filename, Path) and log_filename.exists():
            zip_file.write(log_filename)
    return zip_filename


class ServerSimRunner(threading.Thread):
    """This class maintains updated status of the SimRunner.
    It was decided not to make SimRunner a super class and rather make it manipulate directly the structures of
    SimRunner. The rationale for this, was to avoid confusions between the run() on the Thread class and the
    run on the SimRunner class.
    Making a class derive from two different classes needs to be handled carefully.

    In consequence of the rationale above, many of the functions that were handled by the SimRunner are overriden
    by this class.
    """

    def __init__(self, parallel_sims: int = 4, timeout: float = 600.0, verbose=False,
                 output_folder: Optional[str] = None, simulator=None):
        super().__init__(name="SimManager")
        self.runner = SimRunner(simulator=simulator, parallel_sims=parallel_sims, timeout=timeout,
                                verbose=verbose, output_folder=output_folder)
        self.completed_tasks: list[dict[str, Any]] = []  # This is a list of dictionaries with the information of the completed tasks
        self._stop = False

    def run(self) -> None:
        """
        Implements the SimServer process. Basically it makes the housekeeping of the SimRunner vacuuming the
        completed_tasks structure there and populated an richer version inside this class which also contains
        the zipfile that was created by the callback function.
        This function makes a direct manipulation of the structures of SimRunner. This option is not the most proper
        but, it helps keeping the SimRunner as simple as possible, and the extra complexity associated with the Server
        is done by this class.
        When the server is stopped this process cleans all the files that were not yet cleaned.
        """
        while True:
            self.runner.update_completed()
            while len(self.runner.completed_tasks) > 0:
                task = self.runner.completed_tasks.pop(0)
                zip_filename = task.callback_return
                self.completed_tasks.append({
                    'runno': task.runno,
                    'retcode': task.retcode,
                    'circuit': task.netlist_file,
                    'raw': task.raw_file,
                    'log': task.log_file,
                    'zipfile': zip_filename,
                    'start': task.start_time,
                    'stop': task.stop_time,
                })
                _logger.debug(f"Task {task} is finished")
                _logger.debug(self.completed_tasks[-1])
                _logger.debug(len(self.completed_tasks))

            time.sleep(0.2)
            if self._stop is True:
                break
        self.runner.wait_completion()
        self.runner.cleanup_files()  # Delete things that have been left behind

    def add_simulation(self, netlist: Union[str, Path, BaseEditor], *, timeout: Optional[float] = None) -> int:
        """
        Adding a simulation to the list of simulations to be run. The function will return the runno of the simulation
        or -1 if the simulation could not be started.

        :param netlist: The netlist to be simulated
        :param timeout: The timeout for the simulation
        :return: The runno of the simulation or -1 if the simulation could not be started
        """
        _logger.debug(f"starting Simulation of {netlist}")
        task = self.runner.run(netlist, wait_resource=True, timeout=timeout, callback=zip_files, callback_on_error=True)
        if task is None:
            _logger.error(f"Failed to start task {netlist}")
            return -1
        else:
            _logger.info(f"Started task {netlist} with job_id{task.runno}")
            return task.runno

    def _erase_files_and_info(self, pos):
        task = self.completed_tasks[pos]
        for filename in ('circuit', 'log', 'raw', 'zipfile'):
            if filename in task:
                f = task[filename]
                if f is not None and f.exists():
                    _logger.debug(f"deleting file {f}")
                    try:
                        f.unlink()
                        _logger.debug(f"deleted file {f}")
                    except Exception as e:
                        _logger.error(f"Error deleting file {f}: {e}")
        del self.completed_tasks[pos]

    def erase_files_of_runno(self, runno):
        """Will delete all files related with a completed task. Will also delete information on the completed_tasks
        attribute."""
        _logger.debug(f"deleting files of run {runno}")
        for i, task_info in enumerate(self.completed_tasks):
            if task_info['runno'] == runno:
                self._erase_files_and_info(i)
                break

    def cleanup_completed(self):
        _logger.debug("deleting all files of all completed tasks")
        while len(self.completed_tasks):
            self._erase_files_and_info(0)

    def stop(self):
        _logger.info("stopping...ServerSimRunner")
        self._stop = True

    def running(self):
        return self._stop is False



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        run_task.py
# Purpose:     Class used for a spice tool using a process call
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     23-12-2016
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
"""
Internal classes not to be used directly by the user
"""
__author__ = "Nuno Canto Brum <nuno.brum@gmail.com>"
__copyright__ = "Copyright 2023, Fribourg Switzerland"

from copy import copy
from pathlib import Path
import threading
import time
import traceback
from time import sleep
from typing import Callable, Union, Any, Type, Optional
import logging

from ..editor.updates import Updates, UpdateValueType

from .process_callback import ProcessCallback
from .simulator import Simulator

_logger = logging.getLogger("spicelib.RunTask")

END_LINE_TERM = '\n'


def format_time_difference(time_diff):
    """Formats the time difference in a human-readable format, stripping the hours or minutes if they are zero"""
    seconds_difference = int(time_diff)
    milliseconds = int((time_diff - seconds_difference) * 1000)
    hours, remainder = divmod(seconds_difference, 3600)
    minutes, seconds = divmod(remainder, 60)

    if hours == 0:
        if minutes == 0:
            return f"{int(seconds):02d}.{milliseconds:04d} secs"
        else:
            return f"{int(minutes):02d}:{int(seconds):02d}.{milliseconds:04d}"
    else:
        return f"{int(hours):02d}:{int(minutes):02d}:{int(seconds):02d}.{milliseconds:04d}"


class RunTask(threading.Thread):
    """This is an internal Class and should not be used directly by the User."""

    def __init__(self, simulator: Type[Simulator], runno, netlist_file: Path,
                 callback: Optional[Union[Type[ProcessCallback], Callable[[Optional[Path], Optional[Path]], Any]]],
                 callback_args: Union[dict, None] = None,
                 switches: Any = None, timeout: Union[float, None] = None, verbose: bool = False,
                 cwd: Union[str, Path, None] = None,
                 callback_on_error: bool = False,
                 exe_log: bool = False):

        super().__init__(name=f"RunTask#{runno}")
        self.start_time = None
        self.stop_time = None
        self.verbose = verbose
        self.switches = switches
        self.timeout = timeout  # Thanks to Daniel Phili for implementing this
        self.simulator = simulator
        self.runno = runno
        self.netlist_file = netlist_file
        self.callback = callback
        self.callback_args = callback_args
        self.callback_on_error = callback_on_error
        self.cwd = cwd
        self.retcode = -1  # Signals an error by default
        self.raw_file = None
        self.log_file = None
        self.callback_return = None
        self.exe_log = exe_log
        self.exception_text = None
        self._edits = None

    @property
    def edits(self) -> Optional[Updates]:
        return self._edits

    @edits.setter
    def edits(self, netlist_updates: Updates):
        self._edits = copy(netlist_updates)

    def value(self, reference) -> UpdateValueType:
        if not self._edits:
            return None
        return self._edits.value(reference)

    def print_info(self, logger_fun: Callable[[str], None], message: str):
        message = f"RunTask #{self.runno}:{message}"
        logger_fun(message)
        if self.verbose:
            print(f"{time.asctime()} {logger_fun.__name__}: {message}{END_LINE_TERM}")

    def run(self):
        # Running the Simulation

        self.callback_return = None
        self.raw_file = None
        self.log_file = None
        
        self.start_time = time.time()
        self.print_info(_logger.info, ": Starting simulation %d: %s" % (self.runno, self.netlist_file))
        # start execution
        try:
            self.retcode = self.simulator.run(self.netlist_file.absolute().as_posix(), self.switches,
                                              self.timeout, cwd=self.cwd, exe_log=self.exe_log)
        except Exception as e:
            self.exception_text = f"{e.__class__.__name__}: {e}"
            self.retcode = -2
            self.print_info(_logger.error, f"Simulation Failed. {self.exception_text}")
        self.stop_time = time.time()
        # print simulation time with format HH:MM:SS.mmmmmm

        # Calculate the time difference
        sim_time = format_time_difference(self.stop_time - self.start_time)
        # Format the time difference
        self.log_file = self.netlist_file.with_suffix('.log')
        
        some_error = False

        # Cleanup everything
        if self.retcode == 0:
            self.raw_file = self.netlist_file.with_suffix(self.simulator.raw_extension)
            if self.raw_file.exists() and self.log_file.exists():
                # simulation successful
                self.print_info(_logger.info, "Simulation Successful. Time elapsed: %s" % sim_time)
            else:
                self.print_info(_logger.error, "Simulation Raw file or Log file were not found")
                some_error = True
        else:
            # simulation failed
            some_error = True
            self.print_info(_logger.error, "Simulation Aborted. Time elapsed: %s" % sim_time)
            if self.log_file.exists():
                self.log_file = self.log_file.replace(self.log_file.with_suffix('.fail'))

        # Do I need to use callback?
        if self.callback and (self.callback_on_error or not some_error):
            # If the callback function is defined and callback_on_error is True, call the callback function
            # even if the simulation failed
            if self.callback_args is not None:
                callback_print = ', '.join([f"{key}={value}" for key, value in self.callback_args.items()])
            else:
                callback_print = ''
            self.print_info(_logger.info, f"Simulation Finished. Calling...{self.callback.__name__}(rawfile, logfile{callback_print})")
            try:
                if self.callback_args is not None:
                    return_or_process = self.callback(self.raw_file, self.log_file, **self.callback_args)
                else:
                    return_or_process = self.callback(self.raw_file, self.log_file)
            except Exception:
                error = traceback.format_exc()
                self.print_info(_logger.error, error)
            else:
                if isinstance(return_or_process, ProcessCallback):
                    proc = return_or_process
                    proc.start()
                    self.callback_return = proc.queue.get()
                    proc.join()
                else:
                    self.callback_return = return_or_process
            finally:
                callback_start_time = self.stop_time
                self.stop_time = time.time()
                self.print_info(_logger.info, "Simulation Callback Finished. Time elapsed: %s" % format_time_difference(
                    self.stop_time - callback_start_time))
        else:
            self.print_info(_logger.debug, "Simulation Callback not called.")
            self.callback_return = None

    def get_results(self) -> Union[None, Any, tuple[str, str]]:
        """
        Returns the simulation outputs if the simulation and callback function has already finished.
        If the simulation is not finished, it simply returns None. If no callback function is defined, then
        it returns a tuple with (raw_file, log_file).
        If a callback function is defined, it returns whatever the callback function is returning, unless
        the simulation failed, and `callback_on_error` is False (default), in which case it returns None.

        :returns: Tuple with the path to the raw file and the path to the log file
        :rtype: tuple(str, str) or None
        """
        if self.is_alive() or self.start_time is None:  # running or not yet started
            return None

        if self.callback:
            return self.callback_return  # callback_return is guaranteed to be set correctly by `run()`
        else:
            return self.raw_file, self.log_file

    def wait_results(self) -> Union[Any, tuple[str, str]]:
        """
        Waits for the completion of the task and returns a tuple with the raw and log files.
        
        :returns: Tuple with the path to the raw file and the path to the log file. See get_results() for more details.
        :rtype: tuple(str, str)
        """
        while self.is_alive() or self.start_time is None or self.retcode == -1:
            sleep(0.1)
        return self.get_results()



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        qspice_simulator.py
# Purpose:     Represents QSPICE
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     26-08-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
import sys
import os

from pathlib import Path
from typing import Union, Optional
import logging
from ..sim.simulator import Simulator, run_function, SpiceSimulatorError
import subprocess

_logger = logging.getLogger("spicelib.QSpiceSimulator")


class Qspice(Simulator):
    """Stores the simulator location and command line options and is responsible for generating netlists and running
    simulations."""
    
    raw_extension = '.qraw'
    """:meta private:"""
    
    #
    # In QSPICE all traces have double precision. This means that qraw files are not compatible
    # with LTSPICE
    
    # windows paths (that are also valid for wine)
    # Please note that os.path.expanduser and os.path.join are sensitive to the style of slash.
    # Placed in order of preference. The first to be found will be used.
    _spice_exe_win_paths = ["~/Qspice/QSPICE64.exe",
                            "~/AppData/Local/Programs/Qspice/QSPICE64.exe",
                            "C:/Program Files/QSPICE/QSPICE64.exe"]
    
    # the default lib paths, as used by get_default_library_paths
    _default_lib_paths = ["C:/Program Files/QSPICE",
                          "~/Documents/QSPICE"]

    """Searches on the any usual locations for a simulator"""
    # defaults:
    spice_exe = []
    process_name = None  
        
    if sys.platform == "linux" or sys.platform == "darwin":
        # status mid 2024: Qspice has limited support for running under linux+wine, and none for MacOS+wine
        # TODO: when the situation gets more mature, add support for wine. See LTspice for an example.
        spice_exe = []
        process_name = None
    else:  # Windows (well, also aix, wasi, emscripten,... where it will fail.)
        for exe in _spice_exe_win_paths:
            if exe.startswith("~"):
                # expand here, as I use _spice_exe_win_paths also for linux, and expanding earlier will fail
                exe = os.path.expanduser(exe)
            if os.path.exists(exe):
                spice_exe = [exe]
                break        

    # fall through        
    if len(spice_exe) == 0:
        spice_exe = []
        process_name = None
    else:
        process_name = Simulator.guess_process_name(spice_exe[0])        
        _logger.debug(f"Found Qspice installed in: '{spice_exe}' ")

    qspice_args = {
        '-ASCII'     : ['-ASCII'],  # Use ASCII file format for the output data(.qraw) file.
        '-ascii'     : ['-ASCII'],  # lowercase also works. Not sure about the others, so keep it this way.
        '-binary'    : ['-binary'],  # Use binary file format for the output data(.qraw) file.
        '-BSIM1'    : ['-BSIM1'],  # Use the charge-conserving BSIM1 charge model for MOS1, MOS2, and MOS3.
        '-Meyer'    : ['-Meyer'],  # Use the Meyer Capacitance model for MOS1, MOS2, and MOS3.
        '-o'         : ['-o', '<path>'],  # Specify the name of a file for the console output.
        # '-p'         : ['-p'],  # Take the netlist piped from stdin. >> Not used in this implementation.
        '-ProtectSelections': ['-ProtectSelections', '<path>'],  # Protect sections marked with .prot/.unprot with encryption.
        '-ProtectSubcircuits': ['-ProtectSubcircuits', '<path>'],  # Protect the body of subcircuits with encryption.
        '-r'       : ['-r', '<path>'],  # Specify the name of the output data(.qraw) file.
    }
    """:meta private:"""
    
    _default_run_switches = ['-o']
    

    @classmethod
    def valid_switch(cls, switch: str, path: str = '') -> list:
        """
        Validates a command line switch. The following options are available for QSPICE:
        
        * `-ASCII`: Use ASCII file format for the output data(.qraw) file.
        * `-binary`: Use binary file format for the output data(.qraw) file.
        * `-BSIM1`: Use the charge-conserving BSIM1 charge model for MOS1, MOS2, and MOS3.
        * `-Meyer`: Use the Meyer Capacitance model for MOS1, MOS2, and MOS3.
        * `-ProtectSelections <path>`: Protect sections marked with .prot/.unprot with encryption.
        * `-ProtectSubcircuits <path>`: Protect the body of subcircuits with encryption.
        * `-r <path>`: Specify the name of the output data(.qraw) file.
        
        The following parameters will already be filled in by spicelib, and cannot be set:

        * `-o <path>`: Specify the name of a file for the console output.
        
        :param switch: switch to be added.
        :type switch: str
        :param path: path to the file related to the switch being given.
        :type path: str, optional
        :return: Nothing
        """
        
        # format check
        if switch is None:
            return []
        switch = switch.strip()
        if len(switch) == 0:
            return []
        if switch[0] != '-':
            switch = '-' + switch
        
        # will be set anyway?
        if switch in cls._default_run_switches:
            _logger.info(f"Switch {switch} is already in the default switches")
            return []        
        
        # read from the dictionary
        if switch in cls.qspice_args:
            switches = cls.qspice_args[switch]
            switches = [switch.replace('<path>', path) for switch in switches]
            return switches
        else:
            raise ValueError(f"Invalid Switch '{switch}'")

    @classmethod
    def run(cls, netlist_file: Union[str, Path], cmd_line_switches: Optional[list] = None, timeout: Optional[float] = None,
            stdout=None, stderr=None, cwd: Union[str, Path, None] = None, exe_log: bool = False) -> int:
        """Executes a Qspice simulation run.
        
        A raw file and a log file will be generated, with the same name as the netlist file, 
        but with `.qraw` and `.log` extension. You can however change the raw file name using the `-r` switch.

        :param netlist_file: path to the netlist file
        :type netlist_file: Union[str, pathlib.Path]
        :param cmd_line_switches: additional command line options. Best to have been validated by valid_switch(), defaults to None
        :type cmd_line_switches: list, optional
        :param timeout: If timeout is given, and the process takes too long, a TimeoutExpired exception will be raised, defaults to None
        :type timeout: float, optional
        :param stdout: control redirection of the command's stdout. Valid values are None, subprocess.PIPE, subprocess.DEVNULL, an existing file descriptor (a positive integer), 
            and an existing file object with a valid file descriptor. 
            With the default settings of None, no redirection will occur. Also see `exe_log` for a simpler form of control.
        :type stdout: _FILE, optional
        :param stderr: Like stdout, but affecting the command's error output. Also see `exe_log` for a simpler form of control.
        :type stderr: _FILE, optional
        :param cwd: The current working directory to run the command in. If None, no change will be done of the working directory. This may not work as wanted when using the simulator under wine.
        :type cwd: Union[str, pathlib.Path, None], optional
        :param exe_log: If True, stdout and stderr will be ignored, and the simulator's execution console messages will be written to a log file
            (named ...exe.log) instead of console. This is especially useful when running under wine or when running simultaneous tasks.
        :type exe_log: bool, optional            
        :raises SpiceSimulatorError: when the executable is not found.
        :raises NotImplementedError: when the requested execution is not possible on this platform.
        :return: return code from the process
        :rtype: int
        """
        if not cls.is_available():
            _logger.error("================== ALERT! ====================")
            _logger.error("Unable to find the QSPICE executable.")
            _logger.error("A specific location of the QSPICE can be set")
            _logger.error("using the create_from(<location>) class method")
            _logger.error("==============================================")
            raise SpiceSimulatorError("Simulator executable not found.")

        if cmd_line_switches is None:
            cmd_line_switches = []
        elif isinstance(cmd_line_switches, str):
            cmd_line_switches = [cmd_line_switches]
        netlist_file = Path(netlist_file).absolute()  # need absolute path, as early 2025 qspice has a strange repetition bug
                
        log_file = Path(netlist_file).with_suffix('.log').as_posix()
        cmd_run = cls.spice_exe + ['-o', log_file] + [netlist_file.as_posix()] + cmd_line_switches
        # start execution
        if exe_log:
            log_exe_file = netlist_file.with_suffix('.exe.log')
            with open(log_exe_file, "w") as outfile:
                error = run_function(cmd_run, timeout=timeout, stdout=outfile, stderr=subprocess.STDOUT, cwd=cwd)
        else:        
            error = run_function(cmd_run, timeout=timeout, stdout=stdout, stderr=stderr, cwd=cwd)
        return error



import logging
import dataclasses
import io
import os
import re
from collections import OrderedDict
from pathlib import Path
from typing import Union

import numpy as np

from spicelib import SpiceReadException
from spicelib.editor.base_editor import scan_eng
from spicelib.log.logfile_data import try_convert_value
from spicelib.raw.plot_interface import PlotInterface, MIN_BYTES_IN_FILE
from spicelib.raw.raw_classes import Axis, TraceRead

from spicelib.utils.detect_encoding import detect_encoding, EncodingDetectError

_logger = logging.getLogger("spicelib.PlotData")

dtype_map = {
    'double': 'd',
    'complex': 'complex',
    'real': 'f'
}
inv_dtype_map = {v: k for k, v in dtype_map.items()}


def dtype_from_string(dtype_str: str) -> np.dtype:
    """Converts a string representation of a data type to a numpy dtype.

    :param dtype_str: The string representation of the data type
    :type dtype_str: str
    :return: The numpy dtype
    :rtype: np.dtype
    :raises ValueError: if the string representation is not recognized
    """
    if dtype_str in dtype_map:
        return np.dtype(dtype_map[dtype_str])
    else:
        raise ValueError(f"Invalid data type string: {dtype_str}")


@dataclasses.dataclass
class TraceInfo:
    """Class for holding trace information.
    This class is used to hold the information of a trace, including its name, type, and data.
    """
    name: str
    dtype: str
    var_type: str


def namify(spice_ref: str):
    """Translate from V(0,n01) to V__n01__ and I(R1) to I__R1__"""
    if spice_ref.lower() in ('time', 'frequency'):
        return spice_ref
    matchobj = re.match(r'(V|I|P)\((\w+)\)', spice_ref)
    if matchobj:
        return f'{matchobj.group(1)}__{matchobj.group(2)}__'
    else:
        raise NotImplementedError(f'Unrecognized alias type for alias : "{spice_ref}"')


def get_remaining_bytes(raw_file) -> int:
    current_pos = raw_file.tell()
    raw_file.seek(0, os.SEEK_END)
    bytes_remaining = raw_file.tell() - current_pos
    raw_file.seek(current_pos)
    return bytes_remaining


def raw_detect_encoding(raw_file: io.IOBase, header_offset=0) -> str:
    """Detects the encoding of the RAW file.

    :param raw_file: The file object to read from
    :type raw_file: io.BufferedReader
    :param header_offset: If applicable the offset on the file where the header is located
    :return: The encoding of the RAW file
    :rtype: str
    :raises EncodingDetectError: if the encoding cannot be detected
    """
    # I cannot use detectencoding() here, as that only works on real text files.
    # This may be a partially binary file, where that function breaks.

    raw_file.seek(header_offset)
    raw_line = raw_file.read(6)
    if len(raw_line) < 6:
        raise SpiceReadException("Invalid RAW file. File is too short to determine encoding.")
    try:
        line = raw_line.decode(encoding='utf_8')
    except UnicodeDecodeError:
        line = ''
    if line == 'Title:' or line == '\nTitle':
        answer = 'utf_8'
    else:
        try:
            line = raw_line.decode(encoding='utf_16_le')
        except UnicodeDecodeError:
            line = ''
        if line == 'Tit' or line == '\nTi':
            answer = 'utf_16_le'
        else:
            raise SpiceReadException("Invalid RAW file. Unrecognized encoding.")
    raw_file.seek(header_offset)  # reposition the pointer at the beginning of the header
    return answer


class PlotData(PlotInterface):
    """Class for holding plot data.
    This class is used to hold the data of a plot, including the axis and traces.

    Do not instantiate this class directly, use the ``RawRead`` class instead."""

    def __init__(self, raw_file: io.BufferedReader, raw_filename: Path, plot_nr: int, encoding: str,
                 dialect: Union[str, None], verbose: bool):
        """Initializes the PlotData object and reads the RAW file.

        :nodoc:

        This class is not intended to be documented by Sphinx. See RawRead for the parameters. """
        # Initialize and type the instance variables, for the documentation
        self._raw_filename = raw_filename
        self._plot_nr = plot_nr
        self._verbose = verbose
        self.header: list[str] = []  # List to store the header lines
        self._fpos_header = raw_file.tell()  # File position of the header section used to skip the data when reading the header.
        self._raw_params: OrderedDict = OrderedDict()  # Dictionary to store the raw parameters
        self._raw_params['Filename'] = raw_filename.as_posix()  # Storing the filename as part of the dictionary
        # TODO: use backannotations to store the backannotations
        self._backannotations = []  # Storing backannotations
        self._encoding = encoding if encoding else raw_detect_encoding(raw_file, self._fpos_header)  # Encoding of the RAW file, either 'utf_8' or 'utf_16_le'
        self._raw_type = None  # Type of the RAW file, either 'binary' or 'values' but initialized as None to indicate no data is available
        self._aliases = {}  # QSpice defines aliases for some of the traces that can be computed from other traces.
        self._spice_params = {}  # QSpice stores param values in the .raw file. They may have some usage later for
        self._nPoints = 0  # Number of points in the RAW file
        self._nVariables = 0  # Number of variables in the RAW file
        self._dialect = None  # The dialect of the RAW file, either 'ltspice', 'qspice', 'ngspice' or 'xyce'
        self._trace_info: list[TraceInfo] = []  # List of tuples with the trace name and type
        self._read_traces: dict[str, Union[Axis, TraceRead]] = {}
        self._steps = None
        self._axis = None  # Creating the axis
        self._has_axis = False  # Indicates if the RAW file has an axis.
        self._flags = []

        # mark the file position of the header section
        # self._fpos_header = raw_file.tell()

        # Check how many bytes are still available in the file
        bytes_remaining = get_remaining_bytes(raw_file)
        if self._verbose:
            _logger.debug(f"Plot nr {self._plot_nr}: {bytes_remaining} bytes remaining in file.")

        if bytes_remaining < MIN_BYTES_IN_FILE:
            _logger.warning(f"Plot nr {self._plot_nr}: Not enough bytes remaining in file. The plot does not exist or is incomplete.")
            raise SpiceReadException(f"Invalid RAW file. Plot nr {self._plot_nr}: Not enough bytes remaining in file. The plot does not exist.")

        if self._encoding in ['utf_16_le', 'utf-16']:
            # Detect the encoding of the file
            sz_enc = 2
        else:
            sz_enc = 1

        if self._verbose:
            _logger.debug(f"Plot nr. {self._plot_nr}: Reading the file with encoding: '{self._encoding}'")

        # Read the header section of the RAW file

        line = ""
        while True:
            ch = raw_file.read(sz_enc).decode(encoding=self._encoding, errors='replace')
            if len(ch) == 0:
                # End of file reached
                if self._verbose:
                    _logger.warning(f"Plot nr. {self._plot_nr}: End of file reached while reading the header.")
                # will raise an exception later
                break
            if ch == '\n':
                # read one line. must remove the \r
                # if self._encoding == 'utf_8':  # no idea why utf_16_le would not need that, but this 'if' was here
                line = line.rstrip('\r')
                self.header.append(line)
                if line.lower() in ('binary:', 'values:'):
                    self._raw_type = line.lower()
                    break
                line = ""
            else:
                line += ch
        self._fpos_data = raw_file.tell()  # File position of the data section, used to skip the header when reading the data.

        if not self.has_data:
            # No data found this may be valid in some dialects, but not in others,
            # Xyce can have a text section after the binary section, so we cannot raise an exception here
            # this will have to be handled by the caller.
            if self._verbose:
                _logger.info(f"Invalid RAW file. Plot nr. {self._plot_nr}: Header is incomplete.")
            return

        # computing the aliases.
        for line in self.header:
            if line.startswith('.'):  # This is either a .param or a .alias
                if line.startswith('.param'):
                    # This is a .param line which format as the following pattern ".param temp=27"
                    _, _, line = line.partition('.param')
                    k, _, v = line.partition('=')
                    self._spice_params[k.strip()] = v.strip()
                elif line.startswith('.alias'):
                    # This is a parameter line which format as the following pattern ".alias I(R2) (0.0001mho*V(n01,out))"
                    _, alias, formula = line.split(' ', 3)
                    self._aliases[alias.strip()] = formula.strip()
            else:
                # This is the typical RAW style parameter format <param>: <value>
                k, _, v = line.partition(':')
                if k.lower() == 'variables':
                    break
                self._raw_params[k.strip().title()] = v.strip()  # Store the parameter in the dictionary, in title case
        self._nPoints = int(self._raw_params['No. Points'], 10)
        self._nVariables = int(self._raw_params['No. Variables'], 10)
        if self._nPoints == 0 or self._nVariables == 0:
            raise SpiceReadException(f"Invalid RAW file. Plot nr. {self._plot_nr}: No points or variables found: Points: {self._nPoints}, Variables: {self._nVariables}.")

        self._has_axis = self._raw_params['Plotname'].lower() not in ('operating point', 'transfer function', 'integrated noise')

        # autodetect the dialect. This is not always possible
        autodetected_dialect = None
        if 'Command' in self._raw_params:
            if 'ltspice' in self._raw_params['Command'].lower():
                # Can be auto detected
                # binary types: depends on flag, see below
                autodetected_dialect = 'ltspice'
            if 'qspice' in self._raw_params['Command'].lower():
                # Can be auto detected
                # binary types: always double for time, complex for AC
                # see if I already saw an autodetected dialect
                if dialect is None and autodetected_dialect is not None:
                    _logger.warning(f"Plot nr. {self._plot_nr}: Dialect is ambiguous: '{self._raw_params['Command']}'. Using qspice.")
                autodetected_dialect = 'qspice'
            if 'ngspice' in self._raw_params['Command'].lower():
                # Can only be auto detected from ngspice 44 on, as before there was no "Command:"
                # binary types: always double for time, complex for AC
                # see if I already saw an autodetected dialect
                if dialect is None and autodetected_dialect is not None:
                    _logger.warning(f"Plot nr. {self._plot_nr}: Dialect is ambiguous: '{self._raw_params['Command']}'. Using ngspice.")
                autodetected_dialect = 'ngspice'
            if 'xyce' in self._raw_params['Command'].lower():
                # Cannot be auto detected yet (at least not on 7.9, where there is no "Command:")
                #  Flags: real (for time) and complex (for frequency)
                #  Binary types: always double for time, complex for AC
                #  and potentially a text (csv) section that follows, that can be ignored.
                # see if I already saw an autodetected dialect
                if dialect is None and autodetected_dialect is not None:
                    _logger.warning(f"Plot nr. {self._plot_nr}: Dialect is ambiguous: '{self._raw_params['Command']}'. Using xyce.")
                autodetected_dialect = 'xyce'

        if dialect:
            if autodetected_dialect is not None:
                if dialect != autodetected_dialect:
                    _logger.warning(f"Plot nr. {self._plot_nr}: Dialect specified as {dialect}, but the file seems to be from {autodetected_dialect}. Trying to read it anyway.")
        else:
            # no dialect given. Take the autodetected version
            dialect = autodetected_dialect

        # Do I have something?
        if not dialect:
            raise SpiceReadException(f"Invalid RAW file. Plot nr. {self._plot_nr}: file dialect is not specified and could not be auto detected.")

        # and tell the outside world
        self._dialect = dialect

        # set the specifics per dialect
        always_double = dialect != 'ltspice'  # qspice, ngspice and xyce use doubles for everything outside of AC files
        frequency_double = dialect == 'qspice'  # qspice uses double also for frequency for AC files

        # Compiling the raw file information based on the dialect
        self._steps = None
        self._flags = self._raw_params['Flags'].split()
        if 'complex' in self._raw_params['Flags'].lower() or self._raw_params['Plotname'].lower() == 'ac analysis':
            numerical_type = 'complex'
        else:
            if always_double:  # qspice, ngspice and xyce use doubles for everything outside of AC
                numerical_type = 'double'
            elif "double" in self._raw_params['Flags'].lower():  # LTspice: .options numdgt = 7 sets this flag for double precision
                numerical_type = 'double'
            else:
                numerical_type = 'real'
        i = self.header.index('Variables:')

        # Construct the trace information list so to be used on the numpy structured array
        ivar = 0
        for line in self.header[i + 1:-1]:  # Parse the variable names
            line_elmts = line.lstrip().split('\t')
            if len(line_elmts) < 3:
                raise SpiceReadException(f"Invalid RAW file. Plot nr. {self._plot_nr}: Invalid line in the Variables section: {line}")
            name = line_elmts[1]
            var_type = line_elmts[2]
            if ivar == 0:  # If it has an axis, it should be always read
                if numerical_type == 'real':
                    # only ltspice gets here, in non AC
                    axis_numerical_type = 'double'  # LTSpice uses double for the first variable in .OP
                elif numerical_type == 'complex' and frequency_double:
                    axis_numerical_type = 'double'  # QSPICE uses double for frequency for .AC files
                else:
                    axis_numerical_type = numerical_type
                trace_info = TraceInfo(name, dtype_map[axis_numerical_type], var_type)
            else:
                trace_info = TraceInfo(name, dtype_map[numerical_type], var_type)
            self._trace_info.append(trace_info)
            ivar += 1

        if self._verbose:
            _logger.info(f"Plot nr. {self._plot_nr}: Plot is of type '{self.get_plot_name()}', contains {ivar} "
                         f"traces with {self._nPoints} points, reading {len(self._trace_info)} traces.")

        # Setting the properties in the proper format
        self._raw_params["Variables"] = [var.name for var in self._trace_info]

        # Finally, Check for Step Information
        if "stepped" in self._raw_params["Flags"].lower():
            if self._verbose:
                _logger.debug(f"Plot nr. {self._plot_nr}: RAW file has stepped data.")
            try:
                self._load_step_information(raw_filename)
            except SpiceReadException as err:
                _logger.warning(f"Plot nr. {self._plot_nr}: {str(err)}\nError in auto-detecting steps in '{raw_filename}'")
                if self._has_axis:
                    number_of_steps = 0
                    if self._axis is not None:
                        for v in self._axis.data:
                            if v == self._axis.data[0]:
                                number_of_steps += 1
                else:
                    number_of_steps = self._nPoints
                self._steps = [{'run': i + 1} for i in range(number_of_steps)]

        if self._verbose:
            _logger.info(f"Plot nr. {self._plot_nr}: Plot read successfully.")
        # if the file is binary, we need to move the file pointer to the end of the data section
        if self._raw_type == 'binary:':
            # now move the file pointer to the end of the data section, so that the next plot can be read
            record_size = sum([np.dtype(t.dtype).itemsize for t in self._trace_info])
            raw_file.seek(self._fpos_data + self._nPoints * record_size)
        elif self._raw_type == 'values:':
            # In this case for an ASCII file, it will not be lazy and will read all data.
            if self._verbose:
                _logger.debug(f"Plot nr. {self._plot_nr}: ASCII RAW File")
            self._read_ascii_vector(raw_file)

    def _read_ascii_vector(self, raw_file):
        # Create the traces read vector
        for trace_info in self._trace_info:
            if trace_info.name in self._aliases:
                # This is an alias, the data will be computed later
                continue

            # Create the traces for all the info
            trace_data = np.zeros(self._nPoints, dtype=trace_info.dtype)  # Pre-allocate array for speed

            if trace_info.name == self._trace_info[0].name and self._has_axis:
                # This is the axis trace
                self._axis = Axis(trace_info.name, trace_info.var_type, self._nPoints,
                                  inv_dtype_map[trace_info.dtype], trace_data)
                if self._steps is not None:
                    self._axis.set_steps(self._steps)
                self._read_traces[trace_info.name] = self._axis
            else:
                # Create trace object
                trace = TraceRead(trace_info.name, trace_info.var_type, self._nPoints, self._axis,
                                  inv_dtype_map[trace_info.dtype],
                                  trace_data)
                self._read_traces[trace_info.name] = trace

        # Will start the reading of ASCII Values
        for point in range(self._nPoints):
            var_index = 0
            while var_index < len(self._trace_info):
                line = raw_file.readline().decode(encoding=self._encoding, errors='ignore')
                if len(line) == 0:
                    raise SpiceReadException(
                        f"Invalid RAW file. Plot nr. {self._plot_nr}: Invalid data: end of file encountered too early")
                line = line.strip()
                if len(line) == 0:
                    continue  # skip empty lines
                if var_index == 0:
                    s_point, value = line.split("\t", 1)
                    if point != int(s_point):
                        raise SpiceReadException(
                            f"Invalid RAW file. Plot nr. {self._plot_nr}: Invalid data: point is not in sequence ({point} != {int(s_point)})")

                else:
                    value = line

                var = self._trace_info[var_index]
                if var.dtype == 'complex':
                    v = value.split(',')
                    if len(v) != 2:
                        raise SpiceReadException(
                            f"Invalid RAW file. Plot nr. {self._plot_nr}: Invalid data for trace {var.name}: {value} is not a complex value")
                    self._read_traces[var.name].data[point] = complex(float(v[0]), float(v[1]))
                else:
                    self._read_traces[var.name].data[point] = float(value)
                var_index += 1
        # Remaining empty lines if exist are ignored
        while True:
            cursor = raw_file.tell()
            line = raw_file.readline().strip()
            if len(line) != 0:
                raw_file.seek(cursor)  # go back to the beginning of the line
            else:
                break
        if self._verbose:
            _logger.debug(f"Plot nr {self._plot_nr}: ASCII data read successfully.")

    @property
    def has_data(self) -> bool:
        """Indicates if the plot has data."""
        return self._raw_type is not None

    @property
    def dialect(self) -> Union[str, None]:
        """The dialect of the RAW file, either 'ltspice', 'qspice', 'ngspice' or 'xyce'
        """
        return self._dialect

    @property
    def encoding(self) -> str:
        """The encoding of the RAW file, either 'utf_8' or 'utf_16_le'
        """
        return self._encoding

    @property
    def nVariables(self) -> int:
        """Number of variables in the RAW file
        """
        return self._nVariables

    @property
    def nPoints(self) -> int:
        """Number of points in the RAW file
        """
        return self._nPoints

    @property
    def raw_type(self) -> str:
        """The type of the RAW file, either 'binary:' or 'values:'"""
        return self._raw_type

    @property
    def aliases(self) -> dict[str, str]:
        """QSpice defines aliases for some of the traces that can be computed from other traces.
        """
        return self._aliases

    @property
    def backannotations(self) -> list[str]:
        """List to store the backannotations found in the RAW file header
        """
        return self._backannotations

    @property
    def has_axis(self) -> bool:
        """Indicates if the RAW file has an axis.
        This is True for all RAW file plots except for 'Operating Point', 'Transfer Function', and 'Integrated Noise'.
        """
        return self._has_axis

    @property
    def axis(self) -> Union[Axis, None]:
        """
        .. deprecated:: 1.4.5 Use `get_axis()` method instead.

        The axis of the RAW file, if it exists.
        """
        return self._axis

    @property
    def raw_params(self) -> OrderedDict:
        """
        .. deprecated:: 1.4.5 Use `get_raw_properties()` or `get_raw_property()` method instead.

        Dictionary to store the parameters found in the RAW file header.
        The keys are the parameter names, and the values are the parameter values.
        """
        return self._raw_params

    @property
    def flags(self) -> list[str]:
        """List of Flags that are used in this plot. See :doc:`../varia/raw_file` for details.
        """
        return self._flags

    @property
    def steps(self) -> Union[list[dict[str, int]], None]:
        """List of steps in the RAW file, if it exists.
        If the RAW file does not contain stepped data, this will be None.
        If the RAW file contains stepped data, this will be a list of step numbers.
        """
        return self._steps

    def get_raw_property(self, property_name=None) -> Union[str, dict[str, str]]:
        """
        Get a property. By default, it returns all properties defined in the RAW file.

        :param property_name: name of the property to retrieve. If None, all properties are returned.
        :type property_name: str
        :returns: Property object
        :rtype: str
        :raises: ValueError if the property doesn't exist
        """
        # the property name is case-insensitive, but the keys are stored in title case.
        if property_name is None:
            return self.get_raw_properties()
        elif property_name.title() in self._raw_params.keys():
            return self._raw_params[property_name.title()]
        else:
            raise ValueError("Invalid property. Use %s" % str(self._raw_params.keys()))

    def get_raw_properties(self) -> dict[str, str]:
        """
        Get all raw properties.

        :return: Dictionary of all raw properties
        :rtype: dict[str, str]
        """
        return self._raw_params

    def get_plot_name(self) -> str:
        """
        Returns the type of the plot read from the RAW file. Some examples:

        * 'AC Analysis',
        * 'DC transfer characteristic',
        * 'Operating Point',
        * 'Transient Analysis',
        * 'Transfer Function',
        * 'Noise Spectral Density',
        * 'Frequency Response Analysis',
        * 'Noise Spectral Density Curves',
        * 'Integrated Noise'

        :return: plot name
        :rtype: str
        """
        property_name = "Plotname"
        if property_name in self._raw_params.keys():
            return self._raw_params[property_name]
        else:
            return ""

    def get_trace_names(self) -> list[str]:
        """
        Returns a list of exiting trace names of the RAW file.

        :return: trace names
        :rtype: list[str]
        """
        # parsing the aliases needs to be done before implementing this.
        return [trace.name for trace in self._trace_info] + list(self._aliases.keys())

    def _compute_alias(self, alias: str) -> TraceRead:
        """
        Constants like mho need to be replaced and V(ref1,ref2) need to be replaced by (V(ref1)-V(ref2)) and after
        that the aliases can be computed, using the eval() function.
        """
        formula = self._aliases[alias]
        # converting V(ref1, ref2) to (V(ref1)-V(ref2))
        formula = re.sub(r'V\((\w+),0\)', r'V(\1)', formula)
        formula = re.sub(r'V\(0,(\w+)\)', r'(-V(\1))', formula)
        formula = re.sub(r'V\((\w+),(\w+)\)', r'(V(\1)-V(\2))', formula)
        # find all variables used in the formula
        used_vars = [var.name for var in self._trace_info if var.name in formula]
        # converting V(ref1) to V__ref1__ and I(ref1) to I__ref1__
        formula = re.sub(r'([VIP])\((\w+)\)', r'\1__\2__', formula)

        # removing the mho or other constants ex:  (0.0001mho*V(0,n01)) -> (0.0001*V(0,n01))
        formula = re.sub(r'(\d+)((mho)|(ohm))', r'\1', formula)
        if alias.startswith('I('):
            whattype = 'current'
        elif alias.startswith('V('):
            whattype = 'voltage'
        else:
            raise NotImplementedError(f'Unrecognized alias type for alias : "{alias}"')
        trace = TraceRead(alias, whattype, self._nPoints, self._axis, 'double')
        local_vars = {'pi': 3.1415926536, 'e': 2.7182818285}  # This is the dictionary that will be used to compute the alias
        local_vars.update({name: scan_eng(value) for name, value in self._spice_params.items()})
        self.read_trace_data(used_vars)
        local_vars.update({namify(name): self._read_traces[name].data for name in used_vars})
        try:
            trace.data = eval(formula, local_vars)
        except Exception as err:
            raise RuntimeError(f'Error computing alias "{alias}" with formula "{formula}"') from err
        return trace

    @staticmethod
    def _read_bytes_from_file(raw_file: io.BufferedReader, num_bytes: int) -> bytes:
        """Reads a specific number of bytes from a file.

        :param raw_file: The file object to read from
        :type raw_file: io.BufferedReader
        :param num_bytes: The number of bytes to read
        :type num_bytes: int
        :return: The bytes read from the file
        :rtype: bytes
        :raises SpiceReadException: if the number of bytes read is less than requested
        """
        data = raw_file.read(num_bytes)
        if len(data) < num_bytes:
            raise SpiceReadException("Invalid RAW file. Not enough data in the binary section.")
        return data

    def read_trace_data(self, list_of_traces: list[str]):
        """
        Reads the trace data from the binary section of the RAW file.
        """
        with open(self._raw_filename, 'rb') as raw_file:

            if self._raw_type == "binary:":
                # Will start the reading of binary values

                if "fastaccess" in self._raw_params["Flags"].lower():
                    if self._verbose:
                        _logger.debug(f"{self._raw_filename} Binary RAW file with Fast access")
                    # In Fast Accces mode, the data is already stored contiguously per trace

                    previous_data_size = 0
                    start_index = 0
                    # Always read the axis first, if it exists and was not read before
                    if self._has_axis and self._trace_info and (self._axis is None):
                        trace_info = self._trace_info[0]
                        start_index = 1
                        data_size = self._nPoints * np.dtype(trace_info.dtype).itemsize
                        # read this trace
                        raw_file.seek(self._fpos_data + previous_data_size)  # Move to the beginning of the data section
                        raw_data = self._read_bytes_from_file(raw_file, data_size)
                        assert len(raw_data) == data_size, "Invalid RAW file. Not enough data in the binary section."
                        previous_data_size += data_size
                        data = np.frombuffer(raw_data, dtype=trace_info.dtype)
                        # Now create the axis
                        num_type = inv_dtype_map[trace_info.dtype]
                        trace = Axis(trace_info.name, trace_info.var_type, self._nPoints, num_type, data)
                        self._axis = trace
                        self._read_traces[trace_info.name] = trace
                        if self._verbose:
                            _logger.debug(f"Axis '{trace.name}' read successfully. Data type: {trace_info.dtype}, First 5 values: {trace.data[:5]}")
                        # Set the steps if they exist
                        if self._steps is not None:
                            self._axis.set_steps(self._steps)

                    for trace_info in self._trace_info[start_index:]:
                        data_size = np.dtype(trace_info.dtype).itemsize * self._nPoints
                        if trace_info.name in list_of_traces:
                            # read this trace
                            raw_file.seek(self._fpos_data + previous_data_size)  # Move to the beginning of the data section
                            raw_data = self._read_bytes_from_file(raw_file, data_size)
                            assert len(raw_data) == data_size, "Invalid RAW file. Not enough data in the binary section."
                            data = np.frombuffer(raw_data, dtype=trace_info.dtype)
                            num_type = inv_dtype_map[trace_info.dtype]
                            # Now create the trace
                            trace = TraceRead(trace_info.name, trace_info.var_type, self._nPoints, self._axis, num_type, data)
                            self._read_traces[trace_info.name] = trace
                            if self._verbose:
                                _logger.debug(f"Binary data read successfully. Data shape: {data.shape}")
                        previous_data_size += data_size
                else:
                    # Normal Acccess
                    if self._verbose:
                        _logger.debug(f"{self._raw_filename} Binary RAW file with Normal access")

                    # create the numpy num_type for the structured array
                    read_dtypes = []
                    for i, trace_info in enumerate(self._trace_info):
                        if i == 0 and self._has_axis and (self._axis is None):
                            # Always read the axis first, if it exists and was not read before
                            read_dtypes.append((trace_info.name, trace_info.dtype))
                        elif trace_info.name in list_of_traces:
                            read_dtypes.append((trace_info.name, trace_info.dtype))
                        else:
                            # void num_type, to skip this trace
                            read_dtypes.append((trace_info.name, 'V' + str(np.dtype(trace_info.dtype).itemsize)))

                    if self._verbose:
                        _logger.debug(f"{self._raw_filename} Reading traces: {list_of_traces}")
                    record_size = sum([np.dtype(t.dtype).itemsize for t in self._trace_info])
                    raw_file.seek(self._fpos_data)  # Move to the beginning of the data section
                    raw_data = self._read_bytes_from_file(raw_file, self._nPoints * record_size)  # Move to the beginning of the data section
                    data = np.frombuffer(raw_data, dtype=read_dtypes)
                    if self._verbose:
                        _logger.debug(f"Binary data read successfully. Data shape: {data.shape}")
                    for i, trace_info in enumerate(self._trace_info):
                        if i == 0 and self._has_axis and (self._axis is None):
                            # Now create the axis
                            trace_data = data[trace_info.name]
                            num_type = inv_dtype_map[trace_info.dtype]
                            trace = Axis(trace_info.name, trace_info.var_type, self._nPoints, num_type, trace_data)
                            self._axis = trace
                            self._read_traces[trace_info.name] = trace
                            if self._verbose:
                                _logger.debug(f"Axis '{trace.name}' read successfully. Data type: {trace_info.dtype}, First 5 values: {trace.data[:5]}")
                            # Set the steps if they exist
                            if self._steps is not None:
                                self._axis.set_steps(self._steps)
                        elif trace_info.name in list_of_traces:
                            trace_data = data[trace_info.name]
                            num_type = inv_dtype_map[trace_info.dtype]
                            # Now create the trace
                            trace = TraceRead(trace_info.name, trace_info.var_type, self._nPoints, self._axis, num_type, trace_data)
                            self._read_traces[trace_info.name] = trace
                            if self._verbose:
                                _logger.debug(f"Trace '{trace.name}' read successfully. Data type: {trace_info.dtype}, First 5 values: {trace.data[:5]}")
            elif self._raw_type == "values:":
                pass  # nothing to do here, since the reading of ascii raw files is not lazy
            else:
                raise NotImplementedError("Only binary and value RAW files are supported at the moment.")

    def get_trace(self, trace_ref: Union[str, int]) -> Union[Axis, TraceRead]:
        """
        Retrieves the trace with the requested name (trace_ref).

        :param trace_ref: Name of the trace or the index of the trace
        :type trace_ref: str or int
        :return: An object containing the requested trace
        :rtype: DataSet subclass
        :raises IndexError: When a trace is not found
        """
        if isinstance(trace_ref, str):
            for trace_name in self._read_traces:
                if trace_ref.casefold() == trace_name.casefold():  # The trace names are case-insensitive
                    # assert isinstance(trace, DataSet)
                    return self._read_traces[trace_name]
            # not found as a read trace, see if needs to be read from file
            for trace_info in self._trace_info:
                if trace_ref.casefold() == trace_info.name.casefold():  # The trace names are case-insensitive
                    # need to read it from file
                    self.read_trace_data([trace_info.name])
                    return self._read_traces[trace_info.name]
            for alias in self._aliases:
                if trace_ref.casefold() == alias.casefold():
                    return self._compute_alias(alias)
            raise IndexError(f"{self} doesn't contain trace \"{trace_ref}\"\n"
                             f"Valid traces are {[trc.name for trc in self._trace_info]}")
        else:
            if trace_ref < 0 or trace_ref >= len(self._trace_info):
                raise IndexError(f"Trace index {trace_ref} out of range. Valid range is 0 to {len(self._trace_info)-1}.")
            trace_info = self._trace_info[trace_ref]
            return self.get_trace(trace_info.name)  # Recursion with a string

    def get_wave(self, trace_ref: Union[str, int], step: int = 0) -> np.ndarray:
        """
        Retrieves the trace data with the requested name (trace_ref), optionally providing the step number.

        :param trace_ref: Name of the trace or the index of the trace
        :type trace_ref: str or int
        :param step: Optional parameter specifying which step to retrieve.
        :type step: int
        :return: A numpy array containing the requested waveform.
        :rtype: numpy.array
        :raises IndexError: When a trace is not found
        """
        trace = self.get_trace(trace_ref)
        if isinstance(trace, Axis) or isinstance(trace, TraceRead):
            return trace.get_wave(step)
        else:
            raise IndexError(f"{self} doesn't contain trace \"{trace_ref}\"\n"
                             f"Valid traces are {[trc.name for trc in self._trace_info]}")

    def get_time_axis(self, step: int = 0) -> np.ndarray:
        """
        .. deprecated:: 1.0 Use `get_axis()` method instead.

        This function is equivalent to get_trace('time').get_time_axis(step) instruction.
        It's workaround on a LTSpice issue when using 2nd Order compression, where some values on
        the time trace have a negative value."""
        trace_ref = 'time'
        trace = self.get_trace(trace_ref)
        if isinstance(trace, Axis):
            return trace.get_time_axis(step)
        else:
            raise IndexError(f"{self} doesn't contain trace \"{trace_ref}\"\n"
                             f"Valid traces are {[trc.name for trc in self._trace_info]}")

    def get_axis(self, step: int = 0) -> Union[np.ndarray, list[float]]:
        """
        This function is equivalent to get_trace(0).get_wave(step) instruction.
        It also implements a workaround on a LTSpice issue when using 2nd Order compression, where some values on
        the time trace have a negative value.

        :param step: Step number, defaults to 0
        :type step: int, optional
        :raises RuntimeError: if the RAW file does not have an axis.
        :return: Array with the X axis
        :rtype: Union[np.ndarray, list[float]]
        """
        if self._axis:
            axis = self.get_trace(0)
            assert isinstance(axis, Axis), "This RAW file does not have an axis."
            return axis.get_wave(step)
        else:
            raise RuntimeError("This RAW file does not have an axis.")

    def get_len(self, step: int = 0) -> int:
        """
        Returns the length of the data at the give step index.

        :param step: the step index, defaults to 0
        :type step: int, optional
        :return: The number of data points
        :rtype: int
        """
        if self._axis is None:
            # If there is no axis, the length is the number of points
            return self._nPoints
        else:
            # If there is an axis, the length is the length of the axis at the given step
            assert isinstance(self._axis, Axis), "This RAW file does not have an axis."
            return self._axis.get_len(step)

    def _load_step_information(self, filename: Path):
        if 'Command' not in self._raw_params:
            # probably ngspice before v44 or xyce. And anyway, ngspice does not support the '.step' directive
            # FYI: ngspice can do something like .step via a control section with while loop.
            raise SpiceReadException("Unsupported simulator. Only LTspice and QSPICE are supported.")
        if not isinstance(self._raw_params['Command'], str):
            raise SpiceReadException("Invalid Command parameter. Expected a string.")

        if 'ltspice' in self._raw_params['Command'].lower():
            # look in the .log file for information about the steps
            if filename.suffix != '.raw':
                raise SpiceReadException("Invalid Filename. The file should end with '.raw'")
            # it should have a .log file with the same name
            logfile = filename.with_suffix(".log")
            try:
                encoding = detect_encoding(logfile, r"^((.*\n)?Circuit:|([\s\S]*)--- Expanded Netlist ---)")
                log = open(logfile, 'r', errors='replace', encoding=encoding)
            except OSError:
                raise SpiceReadException("Log file '%s' not found" % logfile)
            except UnicodeError:
                raise SpiceReadException("Unable to parse log file '%s'" % logfile)
            except EncodingDetectError:
                raise SpiceReadException("Unable to parse log file '%s'" % logfile)

            for line in log:
                if line.startswith(".step"):
                    step_dict = {}
                    for tok in line[6:-1].split(' '):
                        if '=' not in tok:
                            continue
                        key, value = tok.split('=')
                        step_dict[key] = try_convert_value(value)
                    if self._steps is None:
                        self._steps = [step_dict]
                    else:
                        self._steps.append(step_dict)
            log.close()

        elif 'qspice' in self._raw_params['Command'].lower():
            # look in the .log file for information about the steps
            if filename.suffix != '.qraw':
                raise SpiceReadException("Invalid Filename. The file should end with '.qraw'")
            # it should have a .log file with the same name
            logfile = filename.with_suffix(".log")
            try:
                log = open(logfile, 'r', errors='replace', encoding='utf-8')
            except OSError:
                raise SpiceReadException("Log file '%s' not found" % logfile)
            except UnicodeError:
                raise SpiceReadException("Unable to parse log file '%s'" % logfile)

            step_regex = re.compile(r"^(\d+) of \d+ steps:\s+\.step (.*)$")

            for line in log:
                match = step_regex.match(line)
                if match:
                    step_dict = {}
                    step = int(match.group(1))
                    stepset = match.group(2)
                    _logger.debug(f"Plot nr. {self._plot_nr}: Found step {step} with stepset {stepset}.")

                    tokens = stepset.strip('\r\n').split(' ')
                    for tok in tokens:
                        key, value = tok.split("=")
                        # Try to convert to int or float
                        step_dict[key] = try_convert_value(value)
                    if self._steps is None:
                        self._steps = [step_dict]
                    else:
                        self._steps.append(step_dict)
            log.close()

        else:
            raise SpiceReadException("Unsupported simulator. Only LTspice and QSPICE are supported.")

    def __getitem__(self, item):
        """Helper function to access traces by using the [ ] operator."""
        return self.get_trace(item)

    def get_steps(self, **kwargs) -> Union[list[int], range]:
        """Returns the steps that correspond to the query set in the `**kwargs` parameters.
        Example: ::

            raw_read.get_steps(V5=1.2, TEMP=25)

        This will return all steps in which the voltage source V5 was set to 1.2V and the TEMP parameter is 24 degrees.
        This feature is only possible if a .log file with the same name as the .raw file exists in the same directory.
        Note: the correspondence between step numbers and .STEP information is stored on the .log file.

        :key kwargs:
         key-value arguments in which the key correspond to a stepped parameter or source name, and the value is the
         stepped value.

        :return: The steps that match the query
        :rtype: list[int]
        """
        if self._steps is None:
            return [0]  # returns a single step
        else:
            if len(kwargs) > 0:
                ret_steps = []  # Initializing an empty array
                i = 0
                for step_dict in self._steps:
                    for key in kwargs:
                        ll = step_dict.get(key, None)
                        if ll is None:
                            break
                        elif kwargs[key] != ll:
                            break
                    else:
                        ret_steps.append(i)  # All the step parameters match
                    i += 1
                return ret_steps
            else:
                return range(len(self._steps))  # Returns all the steps

    def export(self, columns: Union[list, None] = None, step: Union[int, list[int]] = -1, **kwargs) -> dict[str, list]:
        """
        Returns a native python class structure with the requested trace data and steps.
        It consists of an ordered dictionary where the columns are the keys and the values are lists with the data.

        This function is used by the export functions.

        :param step: Step number to retrieve. If not given, it will return all steps
        :type step: int
        :param columns: List of traces to use as columns. Default is all traces
        :type columns: list
        :param kwargs: Additional arguments to pass to the pandas.DataFrame constructor
        :type kwargs: ``**dict``
        :return: A pandas DataFrame
        :rtype: pandas.DataFrame
        """
        if columns is None:
            columns = self.get_trace_names()  # if no columns are given, use all traces
        else:
            if self._axis and self._axis.name not in columns:  # If axis is not in the list, add it
                columns.insert(0, self._axis.name)

        if isinstance(step, list):
            steps_to_read = step  # If a list of steps is given, use it
        elif step == -1:
            steps_to_read = self.get_steps(**kwargs)  # If no step is given, read all steps
        else:
            steps_to_read = [step]  # If a single step is given, pass it as a list

        step_columns = []
        if len(step_columns) > 1:
            if self._steps is not None:
                for step_dict in self._steps[0]:
                    for key in step_dict:
                        step_columns.append(key)
        data = OrderedDict()
        # Create the headers with the column names and empty lists
        for col in columns:
            data[col] = []
        for col in step_columns:
            data[col] = []
        # Read the data
        for step in steps_to_read:
            for col in columns:
                data[col] += list(self.get_wave(col, step))
            if self._steps is not None and step < len(self._steps):
                for col in step_columns:
                    data[col] += [self._steps[step][col]] * len(data[columns[0]])
        return data

    def to_dataframe(self, columns: Union[list, None] = None, step: Union[int, list[int]] = -1, **kwargs):
        """
        Returns a pandas DataFrame with the requested data.

        :param step: Step number to retrieve. If not given, it
        :type step: int
        :param columns: List of traces to use as columns. Default is all traces
        :type columns: list
        :param kwargs: Additional arguments to pass to the pandas.DataFrame constructor
        :type kwargs: ``**dict``
        :return: A pandas DataFrame
        :rtype: pandas.DataFrame
        """
        # cannot type the return values, as pandas is an optional dependency
        try:
            import pandas as pd
        except ImportError:
            raise ImportError("The 'pandas' module is required to use this function.\n"
                              "Use 'pip install pandas' to install it.")
        data = self.export(columns=columns, step=step, **kwargs)
        return pd.DataFrame(data, **kwargs)

    def to_csv(self, filename: Union[str, Path], columns: Union[list[str], None] = None, step: Union[int, list[int]] = -1,
               separator=',', **kwargs):
        """
        Saves the data to a CSV file.

        :param filename: Name of the file to save the data to
        :type filename: str
        :param columns: List of traces to use as columns. Default is all traces
        :type columns: list
        :param step: Step number to retrieve. If not given, it
        :type step: int
        :param separator: separator to use in the CSV file
        :type separator: str
        :param kwargs: Additional arguments to pass to the pandas.DataFrame.to_csv function
        :type kwargs: ``**dict``
        """
        try:
            import pandas as pd
        except ImportError:
            use_pandas = False
        else:
            use_pandas = True

        if use_pandas:
            df = self.to_dataframe(columns=columns, step=step)
            df.to_csv(filename, sep=separator, **kwargs)
        else:
            # Export to CSV using python built-in functions
            data = self.export(columns=columns, step=step)
            with open(filename, 'w') as f:
                f.write(separator.join(data.keys()) + '\n')
                firstcolumn = list(data.keys())[0]
                for i in range(len(data[firstcolumn])):
                    f.write(separator.join([str(data[col][i]) for col in data.keys()]) + '\n')

    def to_excel(self, filename: Union[str, Path], columns: Union[list, None] = None, step: Union[int, list[int]] = -1, **kwargs):
        """
        Saves the data to an Excel file.

        :param filename: Name of the file to save the data to
        :type filename: Union[str, pathlib.Path]
        :param columns: List of traces to use as columns. Default is None, meaning all traces
        :type columns: list, optional
        :param step: Step number to retrieve, defaults to -1
        :type step: Union[int, list[int]], optional
        :param kwargs: Additional arguments to pass to the pandas.DataFrame.to_excel function
        :type kwargs: ``**dict``
        :raises ImportError: when the 'pandas' module is not installed
        """
        try:
            import pandas as pd
        except ImportError:
            raise ImportError("The 'pandas' module is required to use this function.\n"
                              "Use 'pip install pandas' to install it.")
        df = self.to_dataframe(columns=columns, step=step)
        df.to_excel(filename, **kwargs)



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        ltspice_simulator.py
# Purpose:     Represents a LTspice tool and it's command line options
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     23-12-2016
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
import sys
import os

from pathlib import Path
from typing import Union, Optional
import logging
from ..sim.simulator import Simulator, run_function, SpiceSimulatorError
import subprocess

_logger = logging.getLogger("spicelib.LTSpiceSimulator")


class LTspice(Simulator):
    """Stores the simulator location and command line options and is responsible for generating netlists and running
    simulations."""

    """Searches on the any usual locations for a simulator"""
    # windows paths (that are also valid for wine)
    # Please note that os.path.expanduser and os.path.join are sensitive to the style of slash.
    # Placed in order of preference. The first to be found will be used.
    _spice_exe_win_paths = ["~/AppData/Local/Programs/ADI/LTspice/LTspice.exe",
                            "~/Local Settings/Application Data/Programs/ADI/LTspice/LTspice.exe",
                            "C:/Program Files/ADI/LTspice/LTspice.exe",
                            "C:/Program Files/LTC/LTspiceXVII/XVIIx64.exe",
                            "C:/Program Files (x86)/LTC/LTspiceXVII/XVIIx64.exe",
                            "C:/Program Files (x86)/LTC/LTspiceIV/scad3.exe"
                            ]
    
    # the default lib paths, as used by get_default_library_paths
    _default_lib_paths = ["~/AppData/Local/LTspice/lib",
                          "~/Documents/LTspiceXVII/lib/",
                          "~/Documents/LTspice/lib/",
                          "~/My Documents/LTspiceXVII/lib/",
                          "~/My Documents/LTspice/lib/",
                          "~/Local Settings/Application Data/LTspice/lib"]
    
    # defaults:
    spice_exe = []
    process_name = None       
    
    if sys.platform == "linux" or sys.platform == "darwin":
        # Linux: look for wine and ltspice under wine.
        # MacOS: give preference to wine. If not found: look for native LTspice
        
        # Anything specified in environment variables?
        spice_folder = os.environ.get("LTSPICEFOLDER")
        spice_executable = os.environ.get("LTSPICEEXECUTABLE")

        if spice_folder and spice_executable:
            spice_exe = ["wine", os.path.join(spice_folder, spice_executable)]
            process_name = spice_executable
        elif spice_folder:
            spice_exe = ["wine", os.path.join(spice_folder, "/XVIIx64.exe")]
            process_name = "XVIIx64.exe"
        elif spice_executable:
            default_folder = os.path.expanduser(
                "~/.wine/drive_c/Program Files/LTC/LTspiceXVII"
            )
            spice_exe = ["wine", os.path.join(default_folder, spice_executable)]
            process_name = spice_executable
        else:
            # This is still "linux or darwin"
            # no environment variables was given. Do a search.
            for exe in _spice_exe_win_paths:
                # make the file path wine compatible
                # Note that wine also accepts paths like 'C:\users\myuser\...'.
                # BUT, if I do that, I would not be able to check for the presence of the exe.
                # So: expand everything.                 
                # Linux would use this: 
                #    '/home/myuser/.wine/drive_c/users/myuser/AppData/...'  for _spice_exe_win_paths[0]
                # or '/home/myuser/.wine/drive_c/Program Files/...'         for _spice_exe_win_paths[2]
                # MacOS would use this: 
                #    '/Users/myuser/.wine/drive_c/users/myuser/AppData/...' for _spice_exe_win_paths[0]    
                # or '/Users/myuser/.wine/drive_c/Program Files/...'        for _spice_exe_win_paths[2]  
                # Note that in the user path versions (_spice_exe_win_paths[0] and [1]), I have 2 expansions of the user name.
                if exe.startswith("~"):
                    exe = "C:/users/" + os.path.expandvars("${USER}" + exe[1:])
                # Now I have a "windows" path (but with forward slashes). Make it into a path under wine.
                exe = os.path.expanduser(exe.replace("C:/", "~/.wine/drive_c/"))
                
                if os.path.exists(exe):
                    spice_exe = ["wine", exe]
                    # Note that one easy method of killing a wine process is to run "wineserver -k", 
                    # but we kill via psutil....kill(), as that would also fit non-wine executions.
                    
                    break
            else:
                # The else block will not be executed if the loop is stopped by a break statement.
                # in case of MacOS, try the native LTspice as last resort
                if sys.platform == "darwin":
                    exe = '/Applications/LTspice.app/Contents/MacOS/LTspice'
                    if os.path.exists(exe):
                        spice_exe = [exe]
                                
    else:  # Windows (well, also aix, wasi, emscripten,... where it will fail.)
        for exe in _spice_exe_win_paths:
            if exe.startswith("~"):
                # expand here, as I use _spice_exe_win_paths also for linux, and expanding earlier will fail
                exe = os.path.expanduser(exe)
            if os.path.exists(exe):
                spice_exe = [exe]
                break
    
    # The following variables are not needed anymore. This also makes sphinx not mention them in the documentation.
    del exe
    if sys.platform == "linux" or sys.platform == "darwin":
        del spice_folder
        del spice_executable
            
    # fall through        
    if len(spice_exe) == 0:
        spice_exe = []
        process_name = None
    else:
        process_name = Simulator.guess_process_name(spice_exe[0])
        _logger.debug(f"Found LTspice installed in: '{spice_exe}' ")

    ltspice_args = {
        '-alt'                : ['-alt'],  # Set solver to Alternate.
        '-ascii'              : ['-ascii'],  # Use ASCII.raw files. Seriously degrades program performance.
        # 'batch'            : ['-b <path>'], # Used by run command: Run in batch mode.E.g. "ltspice.exe-b deck.cir" will leave the data infile deck.raw
        '-big'                : ['-big'],  # Start as a maximized window.
        '-encrypt'            : ['-encrypt'],
        # Encrypt a model library.For 3rd parties wishing to allow people to use libraries without
        # revealing implementation details. Not used by AnalogDevices models.
        '-fastaccess'         : ['-FastAccess'],  # Batch conversion of a binary.rawfile to Fast Access format.
        '-FixUpSchematicFonts': ['-FixUpSchematicFonts'],
        # Convert the font size field of very old user - authored schematic text to the modern default.
        '-FixUpSymbolFonts'   : ['-FixUpSymbolFonts'],
        # Convert the font size field of very old user - authored symbols to the modern default.
        # See Changelog.txt for application hints.
        '-ini'                : ['- ini', '<path>'],  # Specify an .ini file to use other than %APPDATA%\LTspice.ini
        '-I'                  : ['-I<path>'],  # Specify a path to insert in the symbol and file search paths.
        # Must be the last specified option.
        # No space between "-I" and < path > is allowed.
        '-max'                : ['-max'],  # Synonym for -big
        '-netlist'            : ['-netlist'],  # Batch conversion of a schematic to a netlist.
        '-norm'               : ['-norm'],  # Set solver to Normal.
        '-PCBnetlist': ['-PCBnetlist'],  # Batch conversion of a schematic to a PCB format netlist.
        # 'run'              : ['-Run', '-b', '{path}'],  # Start simulating the schematic opened on the command line without
        # pressing the Run button.
        '-SOI'                : ['-SOI'],  # Allow MOSFET's to have up to 7 nodes even in subcircuit expansion.
        '-sync'               : ['-sync'],  # Update component libraries
        # '-uninstall'          : ['-uninstall'],  # Please don't. Executes one step of the uninstallation process. >> Not used in this implementation.
    }
    """:meta private:"""
    
    _default_run_switches = ['-Run', '-b']    

    @classmethod
    def using_macos_native_sim(cls) -> bool:
        """Tells if the simulator used is the MacOS native LTspice

        :return: True if the MacOS native LTspice is used, False otherwise (will also return False on Windows or Linux)
        :rtype: bool
        """
        return sys.platform == "darwin" and cls.spice_exe and "wine" not in cls.spice_exe[0].lower()

    @classmethod
    def valid_switch(cls, switch: str, path: str = '') -> list:
        """
        Validates a command line switch. The following options are available for Windows/wine LTspice:

        * `-alt`: Set solver to Alternate.
        * `-ascii`: Use ASCII.raw files. Seriously degrades program performance.
        * `-encrypt`: Encrypt a model library.For 3rd parties wishing to allow people to use libraries without revealing implementation details. Not used by AnalogDevices models.
        * `-fastaccess`: Batch conversion of a binary.rawfile to Fast Access format.
        * `-FixUpSchematicFonts`: Convert the font size field of very old user - authored schematic text to the modern default.
        * `-FixUpSymbolFonts`: Convert the font size field of very old user - authored symbols to the modern default. See Changelog.txt for application hints.
        * `-ini <path>`: Specify an .ini file to use other than %APPDATA%\\LTspice.ini
        * `-I<path>`: Specify a path to insert in the symbol and file search paths. Must be the last specified option.
        * `-netlist`: Batch conversion of a schematic to a netlist.
        * `-normal`: Set solver to Normal.
        * `-PCBnetlistBatch`: Conversion of a schematic to a PCB format netlist.
        * `-SOI`: Allow MOSFET's to have up to 7 nodes even in subcircuit expansion.
        * `-sync`: Update component libraries
            
        The following parameters will already be filled in by spicelib, and cannot be set:
        
        * `-Run`: Start simulating the schematic opened on the command line without pressing the Run button.
        * `-b`: Run in batch mode.
                
        MacOS native LTspice accepts no command line switches (yet), use it under wine for full support.

        :param switch: switch to be added.
        :type switch: str
        :param path: path to the file related to the switch being given.
        :type path: str, optional
        :return: Nothing
        """
        
        # See if the MacOS simulator is used. If so, check if I use the native simulator
        if cls.using_macos_native_sim():
            # this is the native LTspice. It has no useful command line switches (except '-b').
            raise ValueError("MacOS native LTspice does not support command line switches. Use it under wine for full support.")
            
        # format check
        if switch is None:
            return []
        switch = switch.strip()
        if len(switch) == 0:
            return []
        if switch[0] != '-':
            switch = '-' + switch
        
        # will be set anyway?
        if switch in cls._default_run_switches:
            _logger.info(f"Switch {switch} is already in the default switches")
            return []        
            
        if switch in cls.ltspice_args:
            switches = cls.ltspice_args[switch]
            switches = [switch.replace('<path>', path) for switch in switches]
            return switches
        else:
            raise ValueError(f"Invalid Switch '{switch}'")

    @classmethod
    def run(cls, netlist_file: Union[str, Path], cmd_line_switches: Optional[list] = None, timeout: Optional[float] = None,
            stdout=None, stderr=None, cwd: Union[str, Path, None] = None, exe_log: bool = False) -> int:
        """Executes a LTspice simulation run.
        
        A raw file and a log file will be generated, with the same name as the netlist file, 
        but with `.raw` and `.log` extension.        

        :param netlist_file: path to the netlist file
        :type netlist_file: Union[str, pathlib.Path]
        :param cmd_line_switches: additional command line options. Best to have been validated by valid_switch(), defaults to None
        :type cmd_line_switches: list, optional
        :param timeout: If timeout is given, and the process takes too long, a TimeoutExpired exception will be raised, defaults to None
        :type timeout: float, optional
        :param stdout: control redirection of the command's stdout. Valid values are None, subprocess.PIPE, subprocess.DEVNULL, an existing file descriptor (a positive integer), 
            and an existing file object with a valid file descriptor. 
            With the default settings of None, no redirection will occur. Also see `exe_log` for a simpler form of control.
        :type stdout: _FILE, optional
        :param stderr: Like stdout, but affecting the command's error output. Also see `exe_log` for a simpler form of control.
        :type stderr: _FILE, optional
        :param cwd: The current working directory to run the command in. If None, no change will be done of the working directory. This may not work as wanted when using the simulator under wine.
        :type cwd: Union[str, pathlib.Path, None], optional
        :param exe_log: If True, stdout and stderr will be ignored, and the simulator's execution console messages will be written to a log file 
            (named ...exe.log) instead of console. This is especially useful when running under wine or when running simultaneous tasks.
        :type exe_log: bool, optional            
        :raises SpiceSimulatorError: when the executable is not found.
        :raises NotImplementedError: when the requested execution is not possible on this platform.
        :return: return code from the process
        :rtype: int
        """
        if not cls.is_available():
            _logger.error("================== ALERT! ====================")
            _logger.error("Unable to find a LTspice executable.")
            _logger.error("A specific location of the LTSPICE can be set")
            _logger.error("using the create_from(<location>) class method")
            _logger.error("==============================================")
            raise SpiceSimulatorError("Simulator executable not found.")

        if cmd_line_switches is None:
            cmd_line_switches = []
        elif isinstance(cmd_line_switches, str):
            cmd_line_switches = [cmd_line_switches]
        netlist_file = Path(netlist_file)
        
        # cannot set raw and log file names or extensions. They are always '<netlist_file>.raw' and '<netlist_file>.log'
        
        if sys.platform == "linux" or sys.platform == "darwin":
            if cls.using_macos_native_sim():
                # native MacOS simulator, which has its limitations
                if netlist_file.suffix.lower().endswith(".asc"):
                    raise NotImplementedError("MacOS native LTspice cannot run simulations on '.asc' files. Simulate '.net' or '.cir' files or use LTspice under wine.")
                
                cmd_run = cls.spice_exe + ['-b'] + [netlist_file.as_posix()] + cmd_line_switches
            else:
                # wine
                # Drive letter 'Z' is the link from wine to the host platform's root directory. 
                # Z: is needed for netlists with absolute paths, but will also work with relative paths.
                cmd_run = cls.spice_exe + ['-Run'] + ['-b'] + ['Z:' + netlist_file.as_posix()] + cmd_line_switches
        else:
            # Windows (well, also aix, wasi, emscripten,... where it will fail.)
            cmd_run = cls.spice_exe + ['-Run'] + ['-b'] + [netlist_file.as_posix()] + cmd_line_switches
        # start execution
        if exe_log:
            log_exe_file = netlist_file.with_suffix('.exe.log')            
            with open(log_exe_file, "w") as outfile:
                error = run_function(cmd_run, timeout=timeout, stdout=outfile, stderr=subprocess.STDOUT, cwd=cwd)
        else:        
            error = run_function(cmd_run, timeout=timeout, stdout=stdout, stderr=stderr, cwd=cwd)
        return error

    @classmethod
    def create_netlist(cls, circuit_file: Union[str, Path], cmd_line_switches: Optional[list] = None, timeout: Optional[float] = None, 
                       stdout=None, stderr=None, 
                       cwd: Union[str, Path, None] = None, exe_log: bool = False) -> Path:
        """Create a netlist out of the circuit file

        :param circuit_file: path to the circuit file
        :type circuit_file: Union[str, pathlib.Path]
        :param cmd_line_switches: additional command line options. Best to have been validated by valid_switch(), defaults to None
        :type cmd_line_switches: list, optional
        :param timeout: If timeout is given, and the process takes too long, a TimeoutExpired exception will be raised, defaults to None
        :type timeout: float, optional        
        :param stdout: control redirection of the command's stdout. Valid values are None, subprocess.PIPE, subprocess.DEVNULL, an existing file descriptor (a positive integer), 
            and an existing file object with a valid file descriptor. 
            With the default settings of None, no redirection will occur. Also see `exe_log` for a simpler form of control.
        :type stdout: _FILE, optional
        :param stderr: Like stdout, but affecting the command's error output. Also see `exe_log` for a simpler form of control.
        :type stderr: _FILE, optional
        :param cwd: The current working directory to run the command in. If None, no change will be done of the working directory. This may not work as wanted when using the simulator under wine.
        :type cwd: Union[str, pathlib.Path, None], optional
        :param exe_log: If True, stdout and stderr will be ignored, and the simulator's execution console messages will be written to a log file 
            (named ...exe.log) instead of console. This is especially useful when running under wine or when running simultaneous tasks.
        :type exe_log: bool, optional            
        :raises NotImplementedError: when the requested execution is not possible on this platform.
        :raises RuntimeError: when the netlist cannot be created
        :return: path to the netlist produced
        :rtype: pathlib.Path
        """
        if not cls.is_available():
            _logger.error("================== ALERT! ====================")
            _logger.error("Unable to find a LTspice executable.")
            _logger.error("A specific location of the LTSPICE can be set")
            _logger.error("using the create_from(<location>) class method")
            _logger.error("==============================================")
            raise SpiceSimulatorError("Simulator executable not found.")
                        
        # prepare instructions, two stages used to enable edits on the netlist w/o open GUI
        # see: https://www.mikrocontroller.net/topic/480647?goto=5965300#5965300
        if cmd_line_switches is None:
            cmd_line_switches = []
        elif isinstance(cmd_line_switches, str):
            cmd_line_switches = [cmd_line_switches]
        circuit_file = Path(circuit_file)
        
        if cls.using_macos_native_sim():
            # native MacOS simulator
            raise NotImplementedError("MacOS native LTspice does not have netlist generation capabilities. Use LTspice under wine.")
        
        cmd_netlist = cls.spice_exe + ['-netlist'] + [circuit_file.as_posix()] + cmd_line_switches
        # start execution
        _logger.info("Creating Netlist for %s" % circuit_file)
        if exe_log:
            log_exe_file = circuit_file.with_suffix('.exe.log')            
            with open(log_exe_file, "w") as outfile:
                error = run_function(cmd_netlist, timeout=timeout, stdout=outfile, stderr=subprocess.STDOUT, cwd=cwd)
        else:
            error = run_function(cmd_netlist, timeout=timeout, stdout=stdout, stderr=stderr, cwd=cwd)

        if error == 0:
            netlist = circuit_file.with_suffix('.net')
            if netlist.exists():
                _logger.debug("OK")
                return netlist
        msg = "Failed to create netlist"
        _logger.error(msg)
        raise RuntimeError(msg)



from tkinter import Tk, TclError
import logging

_logger = logging.getLogger("spicelib.utils.clipboard")


class Clipboard:

    def __init__(self, verbose=False):
        self.verbose = verbose
        self.tk = Tk()
        self.tk.withdraw()

    def __del__(self):
        self.tk.quit()

    def copy(self, data:str) -> None:
        """
        Copy the text contents passed as a paramter onto the OS clipboard 

        :param data: Text to place on the OS clipboard
        :type data: str
        """
        self.clear()
        self.tk.clipboard_append(data)

    def paste(self, clear=False) -> str:
        """
        Return the clipboard contents
        optionally clear the contents

        :param clear: If true, clear the clipboard before returning the data
        :return: Text contents of the clipboard
        :rtype: str
        """
        try:
            contents = self.tk.clipboard_get()
        except TclError as er:
            if self.verbose:
                print(f"TCL error: {er}")
            contents = ""
        if contents is not None and contents != "" and clear:
            self.clear()
        return contents

    def clear(self, check=True) -> bool:
        """
        Clear the contents of the clipboard

        :param check: If true, check the clear was successful
        :return: True if the clipboard is empty after the call or if check is False
        :rtype: bool
        """
        self.tk.clipboard_clear()
        data = self.paste()
        return True if not check else data == ""


    def cut(self, data:str) -> bool:
        """
        As per copy(), but the clipboard is first cleared
        then we return true if successful thus allowing the
        caller to determine if it safe to delete the cut data from memory

        :param data: Text to place on the OS clipboard
        :type data: str
        :return: True if the clipboard is not empty after the call
        :rtype: bool
        """
        # Note we can't reliably compare data and rdata as the OS may
        # alter the encoding or structure between calls

        try:
            self.copy(data)
            rdata = self.paste()
        except self.tk.TclError:
            return False
        
        return rdata is not None and len(rdata) > 0


# Example use
if __name__ == "__main__":
    def test_clipboard():
        cb = Clipboard()
        import datetime
        dt = datetime.datetime.now(datetime.timezone.utc)

        existing = cb.paste()
        if len(existing):
            print(f"Existing clipboard: '{existing}'")
            print("Clearing the clipboard")
            res = cb.clear()
        else:
            print("Clipboard is currently empty")
    
        assert cb.paste() == "", "Failed to clear the clipboard"
        print("...clipboard is empty")
        text = f"Testing 123 @ {dt}"
        print("Trying a cut()")
        cb.cut(text)
        print("Trying a paste()")
        res = cb.paste()
        print("Comparing cut() and paste()")
        assert text == res, f"test fail: {text} != {res}"
        print("Test pass")
        if text == res:
            print("...and the data matched")

    test_clipboard()




#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        sim_runner.py
# Purpose:     Tool used to launch LTSpice simulation in batch mode.
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     23-12-2016
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
"""
Allows launching LTSpice simulations from a Python Script, thus allowing to overcome the 3 dimensions STEP limitation on
LTSpice, update resistor values, or component models.

The code snipped below will simulate a circuit with two different diode models, set the simulation
temperature to 80 degrees, and update the values of R1 and R2 to 3.3k. ::

    from spicelib.sim.sim_runner import SimRunner
    from spicelib.sim.sweep import sweep
    from spicelib.editor.spice_editor import SpiceEditor
    from spicelib.sim.ltspice_simulator import LTspice

    runner = SimRunner(simulator=LTspice, parallel_sims=4)
    editor = SpiceEditor("my_circuit.net")
    editor.set_parameters(temp=80)  # Sets the simulation temperature to be 80 degrees
    editor.set_component_value('R2', '3.3k')  #  Updates the resistor R2 value to be 3.3k
    for dmodel in ("BAT54", "BAT46WJ"):
        editor.set_element_model("D1", model)  # Sets the Diode D1 model
        for res_value in sweep(2.2, 2,4, 0.2):  # Steps from 2.2 to 2.4 with 0.2 increments
            editor.set_component_value('R1', res_value)  #  Updates the resistor R1 value to be 3.3k
            runner.run()

    runner.wait_completion()  # Waits for the LTSpice simulations to complete

    print("Total Simulations: {}".format(runner.runno))
    print("Successful Simulations: {}".format(runner.okSim))
    print("Failed Simulations: {}".format(runner.failSim))

The first line will create a python class instance that represents the LTSpice file or netlist that is to be
simulated. This object implements methods that are used to manipulate the spice netlist. For example, the method
set_parameters() will set or update existing parameters defined in the netlist. The method set_component_value() is
used to update existing component values or models.

---------------
Multiprocessing
---------------

For making better use of today's computer capabilities, the SimRunner spawns several simulation processes
each executing in parallel a simulation.

By default, the number of parallel simulations is 4, however the user can override this in two ways. Either
using the class constructor argument ``parallel_sims`` or by forcing the allocation of more processes in the
run() call by setting ``wait_resource=False``. ::

    `runner.run(wait_resource=False)`

The recommended way is to set the parameter ``parallel_sims`` in the class constructor. ::

    `runner = SimRunner(simulator=LTspice, parallel_sims=8)`

The user then can launch a simulation with the updates done to the netlist by calling the run() method. Since the
processes are not executed right away, but rather just scheduled for simulation, the wait_completion() function is
needed if the user wants to execute code only after the completion of all scheduled simulations.

The usage of wait_completion() is optional. Just note that the script will only end when all the scheduled tasks are
executed.

---------
Callbacks
---------

As seen above, the `wait_completion()` can be used to wait for all the simulations to be finished. However, this is
not efficient from a multiprocessor point of view. Ideally, the post-processing should be also handled while other
simulations are still running. For this purpose, the user can use a function call back.

The callback function is called when the simulation has finished directly by the thread that has handling the
simulation. A function callback receives two arguments.
The RAW file and the LOG file names. Below is an example of a callback function::

    def processing_data(raw_filename, log_filename):
        '''This is a call back function that just prints the filenames'''
        print("Simulation Raw file is %s. The log is %s" % (raw_filename, log_filename)
        # Other code below either using ltsteps.py or raw_read.py
        log_info = LTSpiceLogReader(log_filename)
        rise, measures = log_info.dataset["rise_time"]

The callback function is optional. If  no callback function is given, the thread is terminated just after the
simulation is finished.
"""
__author__ = "Nuno Canto Brum <nuno.brum@gmail.com>"
__copyright__ = "Copyright 2020, Fribourg Switzerland"

__all__ = ['SimRunner', 'SimRunnerTimeoutError', 'AnyRunner', 'ProcessCallback', 'RunTask']

from pathlib import Path
import shutil
import inspect  # Library used to get the arguments of the callback function
import time
from time import sleep, thread_time as clock
from typing import Callable, Union, Type, Protocol, Iterator, Any, Optional
import logging

from .process_callback import ProcessCallback
from ..sim.run_task import RunTask
from ..sim.simulator import Simulator
from ..editor.base_editor import BaseEditor
from spicelib import RawWrite, RawRead
from ..editor.updates import Update


_logger = logging.getLogger("spicelib.SimRunner")
END_LINE_TERM = '\n'


class SimRunnerTimeoutError(TimeoutError):
    """Timeout Error class"""
    ...


IteratorFilterType = Union[Callable[[RunTask], bool], dict, None]
"""This is the type used for filtering RunTasks. See the TaskIterator.conditions parameter documentation."""


class TaskIterator:
    """SimRunner Helper class to iterate tasks. It returns all completed tasks, and if the wait parameter is True,
    it will wait for all the tasks to complete. The conditions parameter can be used to filter tasks that respect a
    given condition. The conditions are specified by the user by a function that receives the task object and
    returns a True equivalent for the task to be returned or a False equivalent when the task is to be rejected.
    In most cases, the user can use the functions task.value(ref) and task.param(name) to compose the conditions
    The return_function parameter specifies what the iterator is returning.

    :param runner: SimRunner class to process
    :type runner: SimRunner
    :param return_function: a callable that receives a RunTask and should return something to the user
    :type return_function: function(task: RunTask) -> Any
    :param conditions: Filter to be used in the iterator. If not given it returns all finished tasks.
        If given, this filter can take two forms:

            * the form of a function that receives a RunTask and returns True or False whether the
            task is included or not. Example condition=lambda x: x.edits[

            * the form of a dictionary where keys are the names of components updated and keys are their
            values. The values can either be a single value, a list, a set or a tuple of values. This means,
            all these possibilities are valid: {'R1': '1k', 'R2':('1k','2k'), 'R3':{'1k', '2k'}, R4:['1k', '2k']}

    :type conditions: None or dict or function(task: RunTask) -> bool
    :param wait: If True, the iterator will wait for the tasks to complete, if False, the iterator only considers
        already completed tasks.
    """

    def __init__(self, runner: "SimRunner", return_function: Callable[[RunTask], Any], wait: bool,
                 conditions: IteratorFilterType = None):
        self.runner = runner
        self.return_function = return_function if return_function is not None else lambda _: True
        self.conditions = conditions
        self.wait = wait
        self._iterator_counter = 0

    def match_conditions(self, runtask: RunTask) -> bool:
        if self.conditions is None:
            return True  # No filter was set
        if not runtask.edits:
            return False   # There are no edits, so a match is not possible
        if isinstance(self.conditions, dict):
            for name, value in self.conditions.items():
                # all conditions must be met. Any failure to match will result in a rejection
                for update in runtask.edits:
                    if isinstance(update, Update):
                        if name == update.name and (
                            value == update.value or (isinstance(value, (list, tuple, set)) and update.value in value)
                        ):
                            break  # force the next iteration of `for name, value in self.conditions.items():`, making it an AND comparison
                    else:
                        # this means: update is a list, but that can never happen, as we do for..in and not a slice
                        # anyway, this construction keeps lint happy
                        continue  # or we should raise a programmer error exception
                else:
                    return False  # No match found.
            return True  # All items on the dictionary were found
        else:
            # if `conditions` is not a dict, then it is a function that will return True or False
            return self.conditions(runtask)

    def __iter__(self):
        self._iterator_counter = 0
        return self

    def __next__(self):
        while True:
            self.runner.update_completed()  # Updates the active_tasks and completed_tasks lists
            # First go through the completed tasks
            if self._iterator_counter < len(self.runner.completed_tasks):
                task: RunTask = self.runner.completed_tasks[self._iterator_counter]
                self._iterator_counter += 1
                if self.match_conditions(task):
                    if task.retcode == 0:
                        return self.return_function(task)
                    else:
                        _logger.error(f"Skipping {task.runno} because simulation failed.")
            else:
                # Then check if there are any active tasks
                if len(self.runner.active_tasks) == 0 or self.wait is False:
                    raise StopIteration

                # Then go through the active tasks to get the maximum timeout
                stop_time = self.runner._maximum_stop_time()

                if stop_time is not None and time.time() > stop_time:  # All tasks are on timeout condition
                    raise SimRunnerTimeoutError(f"Exceeded {self.runner.timeout} seconds waiting for tasks to finish")

                # Wait for the active tasks to finish with a timeout
                sleep(0.2)  # Go asleep for a while


class AnyRunner(Protocol):
    def run(self, netlist: Union[str, Path, BaseEditor], *,
            wait_resource: bool = True,
            callback: Optional[Union[Type[ProcessCallback], Callable]] = None,
            callback_args: Optional[Union[tuple, dict]] = None,
            switches: Optional[dict] = None,
            timeout: Optional[float] = None,
            run_filename: Optional[str] = None,
            exe_log: bool = False) -> Union[RunTask, None]:
        ...

    def wait_completion(self, timeout=None, abort_all_on_timeout=False) -> bool:
        ...

    @property
    def runno(self) -> int:
        """number of total runs"""
        ...

    @property
    def failSim(self) -> int:
        """number of failed simulations"""
        ...

    @property
    def okSim(self) -> int:
        """number of successful completed simulations"""
        ...


class SimRunner(AnyRunner):
    """
    The SimRunner class implements all the methods required for launching batches of Spice simulations.
    
    It is iterable, but with a catch: The iteration will only return the completed tasks (succeeded or not), 
    in the order they were completed. If all completed tasks have been returned, and there are still running tasks, 
    it will wait for the completion of the next task. If you used no callbacks, the result is a tuple with the raw and log file names. 
    If you used callbacks, it will return the return code of the callback function, or None if there was an error. 
    Also see `sim_info()` for more details on the completed tasks.
    
    :param simulator: Forcing a given simulator executable.
    :type simulator: Simulator, optional
    :param parallel_sims: Defines the number of parallel simulations that can be executed at the same time. Ideally this
                          number should be aligned to the number of CPUs (processor cores) available on the machine.
    :type parallel_sims: int, optional
    :param timeout: Timeout parameter as specified on the OS subprocess.run() function. Default is 600 seconds, i.e.
        10 minutes. For no timeout, set to None.
    :type timeout: float, optional
    :param verbose: If True, it enables a richer printout of the program execution.
    :type verbose: bool, optional
    :param output_folder: specifying which directory shall be used for simulation files (raw and log files).
    :type output_folder: str, optional
    :param cwd: The current working directory to run the command in. If None, no change will be done of the working directory.
    :type cwd: str or pathlib.Path, optional

    :raises FileNotFoundError: When the file is not found.  !This will be changed.
    """

    def __init__(self, *, simulator=None, parallel_sims: int = 4, timeout: float = 600.0, verbose=False,
                 output_folder: Union[str, Path, None] = None, cwd: Union[str, Path, None] = None):
        # The '*' in the parameter list forces the user to use named parameters for the rest of the parameters.
        # This is a good practice to avoid confusion.
        self.verbose = verbose
        self.timeout = timeout
        self.cmdline_switches = []

        if output_folder is not None:
            # If not None converts to Path() object
            if not isinstance(output_folder, Path):
                self.output_folder = Path(output_folder)
            else:
                self.output_folder = output_folder
            if not self.output_folder.exists():
                self.output_folder.mkdir()
        else:
            self.output_folder = None

        if cwd is not None:
            # If not None converts to Path() object
            if not isinstance(cwd, Path):
                self.cwd = Path(cwd)
            else:
                self.cwd = cwd
            if not self.cwd.exists():
                self.cwd.mkdir()
        else:
            self.cwd = None

        self.parallel_sims = parallel_sims
        self.active_tasks: list[RunTask] = []
        self.completed_tasks: list[RunTask] = []

        self._runno = 0  # number of total runs
        self._failSim = 0  # number of failed simulations
        self._okSim = 0  # number of successful completed simulations
        # self.failParam = []  # collects for later user investigation of failed parameter sets

        # Gets a simulator.
        if simulator is None:
            raise ValueError("No default simulator defined, please specify a simulator")
        elif issubclass(simulator, Simulator):
            self.simulator = simulator
        else:
            raise TypeError("Invalid simulator type.")
        _logger.info("SimRunner initialized")

    def __del__(self):
        """Class Destructor : Closes Everything"""
        # _logger.debug("Waiting for all spawned sim_tasks to finish.")
        self.wait_completion(abort_all_on_timeout=True)  # Kill all pending simulations
        # _logger.debug("Exiting SimRunner")

    @property
    def runno(self) -> int:
        return self._runno

    @property
    def failSim(self) -> int:
        return self._failSim

    @property
    def okSim(self) -> int:
        return self._okSim

    def sim_info(self) -> dict:
        """
        Returns a dictionary with detailed information of all completed tasks. It is best to be called after the completion of
        all tasks.
        
        The dictionary keys are the run numbers. The values are:
        
            * netlist_file: Path to the netlist file
            * raw_file: Path to the raw file
            * log_file: Path to the log file
            * retcode: Return code of the simulator. -2 means an exception was raised, -1 means the simulation is undefined.            
            * exception_text: Exception information in case of an exception during simulation. None if no exception was raised.
            * callback_return: Return value of the callback function. None if no callback was used.
            * start_time: Start time of the simulation
            * stop_time: Stop time of the simulation
            
        Example: ```{ 1: {'netlist_file': 'circuit1.net', 'raw_file': 'circuit1.raw', 'log_file': 'circuit1.log'```, etc....
            
        :return: Dictionary with detailed information of all completed tasks.
        :rtype: dict
        """
        rv = {}
        for task in self.completed_tasks:
            task: RunTask
            run_no = task.runno
            v = {}
            v['netlist_file'] = task.netlist_file
            v['raw_file'] = task.raw_file
            v['log_file'] = task.log_file
            v['retcode'] = task.retcode
            v['exception_text'] = task.exception_text
            v['callback_return'] = task.callback_return
            v['start_time'] = task.start_time
            v['stop_time'] = task.stop_time
            if task.edits:
                v['edits'] = task.edits.netlist_updates
            rv[run_no] = v
        return rv

    def set_simulator(self, spice_tool: Type[Simulator]) -> None:
        """
        Manually overriding the simulator to be used.

        :param spice_tool: String containing the path to the spice tool to be used, or alternatively the Simulator
            object.
        :type spice_tool: Simulator type
        :return: Nothing
        """
        if issubclass(spice_tool, Simulator):
            self.simulator = spice_tool
        else:
            raise TypeError("Expecting str or Simulator objects")

    def clear_command_line_switches(self):
        """Clear all the command line switches added previously"""
        self.cmdline_switches.clear()

    def add_command_line_switch(self, switch, path=''):
        """
        Used to add an extra command line argument such as '-I<path>' to add symbol search path or '-FastAccess'
        to convert the raw file into Fast Access.
        The argument is a string as is defined in the command line documentation of the used simulator. 
        It is preferred that you use the Simulator's class `valid_switch()` method for validation of the switch.

        :param switch: switch to be added. See Command Line Switches documentation of the used simulator.
        :type switch: str
        :param path: path to the file related to the switch being given.
        :type path: str, optional
        :returns: Nothing
        """
        self.cmdline_switches.append(switch)
        if path is not None:
            self.cmdline_switches.append(path)

    def _on_output_folder(self, afile):
        if self.output_folder:
            return self.output_folder / Path(afile).name
        else:
            return Path(afile)

    def _to_output_folder(self, afile: Path, *, copy: bool, new_name: str = ''):
        if self.output_folder:
            if new_name:
                ddst = self.output_folder / new_name
            else:
                ddst = self.output_folder

            if copy:
                dest = shutil.copy(afile, ddst)
            else:
                dest = shutil.move(afile, ddst)
            return Path(dest)
        else:
            if new_name:
                dest = shutil.copy(afile, afile.parent / new_name)
                return Path(dest)
            else:
                return afile

    def _run_file_name(self, netlist):
        if not isinstance(netlist, Path):
            netlist = Path(netlist)
        if netlist.suffix == '.qsch':
            # The Qsch files can't be simulated, so, they have to be converted to netlist first.
            netlist = netlist.with_suffix('.net')
        return "%s_%i%s" % (netlist.stem, self._runno, netlist.suffix)

    def _prepare_sim(self, netlist: Union[str, Path, BaseEditor], run_filename: Union[str, None]):
        """Internal function"""
        # update number of simulation
        self._runno += 1  # Incrementing internal simulation number
        # Harmonize the netlist into a Path object pointing to a netlist file on the right output folder
        if isinstance(netlist, BaseEditor):
            if run_filename is None:
                run_filename = self._run_file_name(netlist.circuit_file)

            # Calculates the path where to store the new netlist.
            run_netlist_file = self._on_output_folder(run_filename)
            netlist.save_netlist(run_netlist_file)

        elif isinstance(netlist, (Path, str)):
            if run_filename is None:
                run_filename = self._run_file_name(netlist)
            if isinstance(netlist, str):
                netlist = Path(netlist)
            run_netlist_file = self._to_output_folder(netlist, copy=True, new_name=run_filename)
        else:
            raise TypeError("'netlist' parameter shall be a SpiceEditor, Path or a plain str")

        return run_netlist_file

    @staticmethod
    def validate_callback_args(callback: Optional[Callable], callback_args: Union[tuple, dict, None]) -> Union[dict, None]:
        """
        It validates that the callback_args are matching the callback function.
        Note that the first two parameters of the callback functions need to be the raw and log files.

        """
        if callback is None:
            return None  # No callback function, hence callback_args have no effect
        if inspect.isclass(callback) and issubclass(callback, ProcessCallback):
            args = inspect.signature(callback.callback).parameters
        else:
            args = inspect.signature(callback).parameters
        if len(args) < 2:
            raise ValueError("Callback function must have at least two arguments")
        if len(args) > 2:
            if callback_args is None:
                raise ValueError("Callback function has more than two arguments, but no callback_args are given")
            if isinstance(callback_args, dict):
                for pos, param in enumerate(args):
                    if pos > 1:
                        if param not in callback_args:
                            raise ValueError("Callback argument '%s' not found in callback_args" % param)

            if len(args) - 2 != len(callback_args):
                raise ValueError("Callback function has %d arguments, but %d callback_args are given" %
                                 (len(args), len(callback_args))
                                 )
            if isinstance(callback_args, tuple):
                # Convert into a dictionary
                return {param: callback_args[pos - 2] for pos, param in enumerate(args) if pos > 1}
            else:
                return callback_args

    def run(self, netlist: Union[str, Path, BaseEditor], *,
            wait_resource: bool = True,
            callback: Union[Type[ProcessCallback], Callable, None] = None,
            callback_args: Union[tuple, dict, None] = None,
            switches=None,
            timeout: Union[float, None] = None,
            run_filename: Union[str, None] = None,
            callback_on_error: bool = False,
            exe_log: bool = False) -> Union[RunTask, None]:
        """
        Executes a simulation run with the conditions set by the user.
        Conditions are set by the set_parameter, set_component_value or add_instruction functions.

        :param netlist:
            The name of the netlist can be optionally overridden if the user wants to have a better control of how the
            simulations files are generated.
        :type netlist: SpiceEditor or a pathlib.Path to the file
        :param wait_resource:
            Setting this parameter to False will force the simulation to start immediately, irrespective of the number
            of simulations already active.
            By default, the SimRunner class uses only four processors. This number can be overridden by setting
            the parameter ´parallel_sims´ to a different number.
            If there are more than ´parallel_sims´ simulations being done, the new one will be placed on hold till one
            of the other simulations are finished.
        :type wait_resource: bool, optional
        :param callback:
            The user can optionally give a callback function for when the simulation finishes so that processing can
            be done immediately. The callback can either be a function or a class derived from ProcessCallback.
            A callback function must receive two at least input parameters that correspond the
            raw and log files created by the simulation. These need to be the first two parameters of the callback
            function. The other parameters are passed as a dictionary or a tuple in the callback_args parameter.
            If the callback is a class derived from ProcessCallback, then the callback is executed in a separate
            process. The callback function must be defined in the callback() method of the class. As for the callback
            function, the first two parameters are the raw and log files. The other parameters are passed as dictionary
            in the callback_args parameter.

        :type callback: function(raw_file: pathlib.Path, log_file: pathlib.Path, ...), optional
        :param callback_args:
            The callback function arguments. This parameter is passed as keyword arguments to the callback function.
        :type callback_args: dict or tuple, optional
        :param switches: Command line switches override
        :type switches: list
        :param timeout:
            Timeout to be used in waiting for resources. Default time is value defined in this class constructor.
        :type timeout: float, optional
        :param run_filename: Name to be used for the log and raw file.
        :type run_filename: str or pathlib.Path
        :param callback_on_error: If False (default), the callback function is not called if the simulation fails.
            If True, the callback function is called even if the simulation fails.
            Know that in that case it is not guaranteed that the raw and log files will be available.
        :type callback_on_error: bool, optional
        :param exe_log: If True, the simulator's execution console messages will be written to a log file
            (named ...exe.log) instead of console. This is especially useful when running under wine or when running
            simultaneous tasks.
        :type exe_log: bool, optional        
        :returns: The task object of type RunTask. For internal use only.
        :rtype: RunTask
        """
        callback_kwargs = self.validate_callback_args(callback, callback_args)
        if switches is None:
            switches = []
        run_netlist_file = self._prepare_sim(netlist, run_filename)

        if timeout is None:
            timeout = self.timeout

        t0 = clock()  # Store the time for timeout calculation
        while clock() - t0 < timeout + 1:  # Give one second slack in relation to the task timeout
            cmdline_switches = switches or self.cmdline_switches  # If switches are passed, they override the ones
            # inside the class.

            if (wait_resource is False) or (self.active_threads() < self.parallel_sims):
                t = RunTask(
                    simulator=self.simulator, runno=self._runno, netlist_file=run_netlist_file,
                    callback=callback, callback_args=callback_kwargs,
                    switches=cmdline_switches, timeout=timeout, verbose=self.verbose,
                    cwd=self.cwd, callback_on_error=callback_on_error, exe_log=exe_log
                )
                if isinstance(netlist, BaseEditor) and netlist.netlist_updates is not None:
                    t.edits = netlist.netlist_updates  # Copy is made in this assignment
                self.active_tasks.append(t)
                t.start()
                sleep(0.01)  # Give slack for the thread to start
                return t  # Returns the task object
            sleep(0.1)  # Give Time for other simulations to end
        else:
            _logger.error("Timeout waiting for resources for simulation %d" % self._runno)
            if self.verbose:
                _logger.warning("Timeout on launching simulation %d." % self._runno)
            return None

    def run_now(self, netlist: Union[str, Path, BaseEditor], *, switches=None, run_filename: Union[str, None] = None,
                timeout: Union[float, None] = None, exe_log: bool = False) -> tuple[Union[Path, None], Union[Path, None]]:
        """
        Executes a simulation run with the conditions set by the user.
        Conditions are set by the `set_parameter`, `set_component_value` or `add_instruction functions`.

        :param netlist:
            The name of the netlist can be optionally overridden if the user wants to have a better control of how the
            simulations files are generated.
        :type netlist: SpiceEditor or a pathlib.Path to the file
        :param switches: Command line switches override
        :type switches: list
        :param run_filename: Name to be used for the log and raw file.
        :type run_filename: str or pathlib.Path
        :param timeout: Timeout to be used in waiting for resources. Default time is value defined in this class
            constructor.
        :type timeout: float, optional
        :param exe_log: If True, the simulator's execution console messages will be written to a log file 
            (named ...exe.log) instead of console. This is especially useful when running under wine or when running simultaneous tasks.
        :type exe_log: bool, optional
        :returns: the raw and log filenames
        """
        if switches is None:
            switches = []
        run_netlist_file = self._prepare_sim(netlist, run_filename)

        cmdline_switches = switches or self.cmdline_switches  # If switches are passed, they override the ones inside
        # the class.

        if timeout is None:
            timeout = self.timeout

        def dummy_callback(raw, log):
            """Dummy call back that does nothing"""
            return None

        t = RunTask(
            simulator=self.simulator, runno=self._runno, netlist_file=run_netlist_file,
            callback=dummy_callback, callback_args=None,
            switches=cmdline_switches, timeout=timeout, verbose=self.verbose,
            cwd=self.cwd, callback_on_error=False, exe_log=exe_log
        )
        if isinstance(netlist, BaseEditor) and netlist.netlist_updates is not None:
            t.edits = netlist.netlist_updates  # Copy is made in this assignment
        t.start()
        sleep(0.01)  # Give slack for the thread to start
        t.join(timeout + 1)  # Give one second slack in relation to the task timeout
        self.completed_tasks.append(t)
        if t.retcode == 0:
            self._okSim += 1
        else:
            # simulation failed
            self._failSim += 1
        return t.raw_file, t.log_file  # Returns the raw and log file

    def active_threads(self):
        """Returns the number of active simulation runs"""
        self.update_completed()
        return len(self.active_tasks)

    def update_completed(self):
        """
        This function updates the `active_tasks` and `completed_tasks` lists. It moves the finished task from the
        `active_tasks` list to the `completed_tasks` list.
        It should be called periodically to update the status of the simulations.

        :returns: Nothing
        :meta private:
        """
        i = 0
        while i < len(self.active_tasks):
            if self.active_tasks[i].is_alive() or self.active_tasks[i].start_time is None:  # running or not yet started
                i += 1
            else:
                if self.active_tasks[i].retcode == 0:
                    self._okSim += 1
                else:
                    # simulation failed
                    self._failSim += 1
                task = self.active_tasks.pop(i)
                self.completed_tasks.append(task)

    def kill_all_ltspice(self):
        """
        .. deprecated:: 1.0 Use `kill_all_spice()` instead.
        
        This is only here for compatibility with previous code.
        
        Function to terminate LTSpice"""
        self.kill_all_spice()

    def kill_all_spice(self):
        """Function to terminate xxSpice processes"""
        simulator = Simulator
        process_name = simulator.process_name
        import psutil
        for proc in psutil.process_iter():
            # check whether the process name matches

            if proc.name() == process_name:
                _logger.info("killing Spice", proc.pid)
                proc.kill()

    def _maximum_stop_time(self):
        """
        This function will return the maximum timeout time of all active tasks.
        :return: Maximum timeout time or None, if there is no timeout defined.
        :rtype: float or None
        """
        alarm = None
        for task in self.active_tasks:
            tout = task.timeout if task.timeout is not None else self.timeout
            if tout is not None and task.start_time is not None:
                stop = task.start_time + tout
                if alarm is None:
                    alarm = stop
                elif stop > alarm:
                    alarm = stop
        return alarm

    def wait_completion(self, timeout=None, abort_all_on_timeout=False) -> bool:
        """
        This function will wait for the execution of all scheduled simulations to complete.

        :param timeout: Cancels the wait after the number of seconds specified by the timeout.
            This timeout is reset everytime that a simulation is completed. The difference between this timeout and the
            one defined in the SimRunner instance, is that the latter is implemented by the subprocess class, and
            this one just cancels the wait.
        :type timeout: int
        :param abort_all_on_timeout: attempts to stop all LTSpice processes if timeout is expired.
        :type abort_all_on_timeout: bool
        :returns: True if all simulations were executed successfully
        :rtype: bool
        """
        self.update_completed()
        if timeout is not None:
            stop_time = time.time() + timeout
        else:
            stop_time = None
        while len(self.active_tasks) > 0:
            sleep(1)
            self.update_completed()
            if timeout is None:
                stop_time = self._maximum_stop_time()
            if stop_time is not None:  # This can happen if timeout was set as none everywhere
                if time.time() > stop_time:
                    if abort_all_on_timeout:
                        self.kill_all_spice()
                    return False

        return self._failSim == 0

    @staticmethod
    def _del_file_if_exists(workfile: Optional[Path]):
        """
        Deletes a file if it exists.
        :param workfile: File to be deleted
        :type workfile: pathlib.Path
        :return: Nothing
        """
        if workfile is not None and workfile.exists():
            _logger.info("Deleting..." + workfile.name)
            workfile.unlink()

    @staticmethod
    def _del_file_ext_if_exists(workfile: Path, ext: str):
        """
        Deletes a file extension if it exists.
        :param workfile: File to be deleted
        :type workfile: pathlib.Path
        :param ext: Extension to be deleted
        :type ext: str
        :return: Nothing
        """
        sim_file = workfile.with_suffix(ext)
        SimRunner._del_file_if_exists(sim_file)

    def cleanup_files(self):
        """
        Will delete all log and raw files that were created by the script. This should only be executed at the end
        of data processing.
        """
        self.update_completed()  # Updates the active_tasks and completed_tasks lists

        for task in self.completed_tasks:
            netlistfile = task.netlist_file
            self._del_file_if_exists(netlistfile)  # Delete the netlist file if still exists
            self._del_file_if_exists(task.log_file)  # Delete the log file if was created
            self._del_file_if_exists(netlistfile.with_suffix('.exe.log'))  # Delete the log file if was created
            self._del_file_if_exists(task.raw_file)  # Delete the raw file if was created

            if netlistfile.suffix == '.net' or netlistfile.suffix == '.asc':
                # Delete the files that have been potentially created by LTSpice
                for ext in ('.log.raw', '.op.raw'):
                    self._del_file_ext_if_exists(netlistfile, ext)

                if netlistfile.suffix == '.asc':  # If simulated from an asc file, delete the .net file
                    # Then needs to delete the .net as well
                    self._del_file_ext_if_exists(netlistfile, '.net')

    def file_cleanup(self):
        """
        .. deprecated:: 1.0 Use `cleanup_files()` instead.
        """
        self.cleanup_files()  # Alias for backward compatibility, this will be deleted in the future

    # ############ Iterator methods
    # def __len__(self):
    #    return len(self.completed_tasks)

    def __iter__(self):
        """Legacy Iterator, returns the get_results() from the task"""
        return TaskIterator(self, lambda x: x.get_results(), True, None)

    def tasks(self, conditions: IteratorFilterType = None) -> Iterator[RunTask]:
        """
        Returns an iterator which iterates all completed tasks

        :param conditions: Filter to be used in the iterator. See TaskIterator conditions parameter documentation.
        :type conditions: Optional or dict or func(RunTask) -> bool
        :return: Iterator[RunTask]
        """
        return TaskIterator(self, lambda x: x, True, conditions)

    def create_raw_file_with(self, raw_filename: Union[Path, str], save: list[str],
                             conditions: IteratorFilterType) -> bool:
        """
        Creates a new raw_file, with traces belonging to different runs. The type of the raw file is the same as
        the first raw file that is matching the conditions. See filter_completed_tasks() method.

        :param raw_filename: The new RAW filename
        :type raw_filename: str or pathlib.Path
        :param save: A list with traces that are going to be saved in the new raw file
        :type save: list[str]
        :param conditions: A filter as specified on the TaskIterator class
        :type conditions: dict
        :return: True if the raw file was created successfully and includes ALL runs, False otherwise
        :rtype: bool
        """
        
        retval = True

        # Obtain a first task
        first_task: RunTask = next(self.tasks(conditions))

        # Initialize a raw file based on the contents of the first raw
        if not first_task.raw_file:
            _logger.error(f"Raw file {first_task.raw_file} does not exist. Cannot create new raw file.")
            return False
            
        template = RawRead(first_task.raw_file)

        new_raw = RawWrite(
            template.raw_params['Title'],
            fastacces=True,
            numtype='auto',
            encoding='utf_16_le'
        )

        # Go through the tasks that match the conditions given
        for run_task in self.tasks(conditions):
            # Open the raw file
            if not run_task.raw_file:
                _logger.error(f"Raw file {run_task.raw_file} does not exist. Skipping this task.")
                retval = False
                continue
            source_raw = RawRead(run_task.raw_file, traces_to_read=save)
            new_raw.add_traces_from_raw(source_raw, save, force_axis_alignment=True, add_tag=run_task.runno)

        new_raw.save(raw_filename)
        return retval

    def export_sim_log(self, logfile: Union[Path, str]):
        """Exports the simulation log to a file.

        :param logfile: The path to the log file
        :type logfile: Union[Path, str]
        """
        import pprint
        logfile = self._on_output_folder(logfile)
        with open(logfile, 'w') as log_file:
            pprint.pprint(self.sim_info(), log_file)



#!/usr/bin/env python

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        sim_stepping.py
# Purpose:     Spice Simulation Library intended to automate the exploring of
#              design corners, try different models and different parameter
#              settings.
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     31-07-2020
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------

__author__ = "Nuno Canto Brum <nuno.brum@gmail.com>"
__copyright__ = "Copyright 2017, Fribourg Switzerland"

from pathlib import Path
from typing import Callable, Union, Type, Iterable
from functools import wraps
import logging

from spicelib.sim.process_callback import ProcessCallback

_logger = logging.getLogger("spicelib.SimStepper")
from ..editor.base_editor import BaseEditor
from .sim_runner import AnyRunner


class StepInfo(object):
    def __init__(self, what: str, elem: str, iterable: Iterable):
        self.what = what
        self.elem = elem
        self.iter = iterable

    def __len__(self):
        return len(list(self.iter))

    def __str__(self):
        return f"Iteration on {self.what} {self.elem} : {self.iter}"


class SimStepper(AnyRunner):
    """This class is intended to be used for simulations with many parameter sweeps. This provides a more
    user-friendly interface than the SpiceEditor/SimRunner class when there are many parameters to be stepped.

    Using the SpiceEditor/SimRunner classes a loop needs to be added for each dimension of the simulations.
    A typical usage would be as follows: 
    
    .. code-block:: python
    
        netlist = SpiceEditor("my_circuit.asc")
        runner = SimRunner(parallel_sims=4)
        for dmodel in ("BAT54", "BAT46WJ")
            netlist.set_element_model("D1", model)  # Sets the Diode D1 model
            for res_value1 in sweep(2.2, 2,4, 0.2):  # Steps from 2.2 to 2.4 with 0.2 increments
                netlist.set_component_value('R1', res_value1)  # Updates the resistor R1 value to be 3.3k
                for temperature in sweep(0, 80, 20):  # Makes temperature step from 0 to 80 degrees in 20 degree steps
                    netlist.set_parameters(temp=80)  # Sets the simulation temperature to be 80 degrees
                    for res_value2 in (10, 25, 32):
                        netlist.set_component_value('R2', res_value2)  # Updates the resistor R2 value to be 3.3k
                        runner.run(netlist)

        runner.wait_completion()  # Waits for the Spice simulations to complete

    With SimStepper the same thing can be done as follows, resulting in a cleaner code:

    .. code-block:: python

        netlist = SpiceEditor("my_circuit.asc")
        Stepper = SimStepper(netlist, SimRunner(parallel_sims=4, output_folder="./output"))
        Stepper.add_model_sweep('D1', "BAT54", "BAT46WJ")
        Stepper.add_component_sweep('R1', sweep(2.2, 2,4, 0.2))  # Steps from 2.2 to 2.4 with 0.2 increments
        Stepper.add_parameter_sweep('temp', sweep(0, 80, 20))  # Makes temperature step from 0 to 80 degrees in 20
                                                            # degree steps
        Stepper.add_component_sweep('R2', (10, 25, 32)) #  Updates the resistor R2 value to be 3.3k
        Stepper.run_all()

    Another advantage of using SimStepper is that it can optionally use the .SAVEBIAS in the first simulation and
    then use the .LOADBIAS command at the subsequent ones to speed up the simulation times.
    """

    def __init__(self, circuit: BaseEditor, runner: AnyRunner):
        self.runner = runner
        self.netlist = circuit
        self.iter_list: list[StepInfo] = []
        self.current_values = {}
        self.sim_info = {}

    @wraps(BaseEditor.add_instruction, updated=())  # updated=() solves conflict between wraps and abstract classes
    def add_instruction(self, instruction: str):
        self.netlist.add_instruction(instruction)

    @wraps(BaseEditor.add_instructions, updated=())  # updated=() solves conflict between wraps and abstract classes
    def add_instructions(self, *instructions) -> None:
        self.netlist.add_instructions(*instructions)

    @wraps(BaseEditor.remove_instruction, updated=())  # updated=() solves conflict between wraps and abstract classes
    def remove_instruction(self, instruction) -> bool:
        return self.netlist.remove_instruction(instruction)

    @wraps(BaseEditor.remove_Xinstruction, updated=())  # updated=() solves conflict between wraps and abstract classes
    def remove_Xinstruction(self, search_pattern: str) -> bool:
        return self.netlist.remove_Xinstruction(search_pattern)

    @wraps(BaseEditor.set_parameters, updated=())  # updated=() solves conflict between wraps and abstract classes
    def set_parameters(self, **kwargs):
        self.netlist.set_parameters(**kwargs)
        self.current_values.update(**kwargs)

    @wraps(BaseEditor.set_parameter, updated=())  # updated=() solves conflict between wraps and abstract classes
    def set_parameter(self, param: str, value: Union[str, int, float]) -> None:
        self.netlist.set_parameter(param, value)
        self.current_values[param] = value

    @wraps(BaseEditor.set_component_values, updated=())  # updated=() solves conflict between wraps and abstract classes
    def set_component_values(self, **kwargs):
        self.netlist.set_component_values(**kwargs)
        self.current_values.update(**kwargs)

    @wraps(BaseEditor.set_component_value, updated=())  # updated=() solves conflict between wraps and abstract classes
    def set_component_value(self, device: str, value: Union[str, int, float]) -> None:
        self.netlist.set_component_value(device, value)
        self.current_values[device] = value

    @wraps(BaseEditor.set_element_model, updated=())  # updated=() solves conflict between wraps and abstract classes
    def set_element_model(self, element: str, model: str) -> None:
        self.netlist.set_element_model(element, model)
        self.current_values[element] = model

    def add_param_sweep(self, param: str, iterable: Iterable):
        """Adds a dimension to the simulation, where the param is swept."""
        self.iter_list.append(StepInfo("param", param, iterable))

    def add_value_sweep(self, comp: str, iterable: Iterable):
        """Adds a dimension to the simulation, where a component value is swept."""
        # The next line raises an ComponentNotFoundError if the component doesn't exist
        _ = self.netlist.get_component_value(comp)
        self.iter_list.append(StepInfo("component", comp, iterable))

    def add_model_sweep(self, comp: str, iterable: Iterable):
        """Adds a dimension to the simulation, where a component model is swept."""
        # The next line raises an ComponentNotFoundError if the component doesn't exist
        _ = self.netlist.get_component_value(comp)
        self.iter_list.append(StepInfo("model", comp, iterable))

    def total_number_of_simulations(self):
        """Returns the total number of simulations foreseen."""
        total = 1
        for step in self.iter_list:
            l = len(step)
            if l:
                total *= l
            else:
                _logger.debug(f"'{step}' is empty.")
        return total

    def run_all(self,
                callback: Union[Type[ProcessCallback], Callable] = None,
                callback_args: Union[tuple, dict] = None,
                switches=None,
                timeout: float = None,
                wait_completion: bool = True,
                filenamer: Callable[[dict[str, str]], str] = None,
                exe_log: bool = False,
                ) -> None:
        """
        Runs all sweeps configured with the methods:

            - add_value_sweep()
            - add_model_sweep()
            - add_param_sweep()

        This function will call the SimRunner run method for each combination of the sweeps defined.
        The parameters are mostly the same as in the SimRunner.run() method, except the filenamer and
        wait_completion parameters.

        :param callback: See the SimRunner run method.
        :type callback: function(raw_file: pathlib.Path, log_file: pathlib.Path, ...), optional
        :param callback_args: See the SimRunner run method.
        :type callback_args: dict or tuple, optional
        :param switches: Command line switches override
        :type switches: list
        :param timeout: See the SimRunner run method.
        :type timeout: float, optional
        :param wait_completion:  See the SimRunner run method.
        :type wait_completion: bool, optional
        :param filenamer:
            A function that receives a dictionary in keyword form (``**dict``) and returns a string. This string will be
            passed to the run_filename parameter on the SimRunner run method. It is important that the function assures
            a unique filename per simulation.
        :type filenamer: Callable receiving keyword parameters.
        :param exe_log: See the SimRunner run method.
        :type exe_log: bool, optional
        :returns: Nothing
        """
        iter_no = 0
        iterators = [iter(step.iter) for step in self.iter_list]
        while True:
            while 0 <= iter_no < len(self.iter_list):
                try:
                    value = iterators[iter_no].__next__()
                except StopIteration:
                    iterators[iter_no] = iter(self.iter_list[iter_no].iter)
                    iter_no -= 1
                    continue

                self.current_values[self.iter_list[iter_no].elem] = value
                if self.iter_list[iter_no].what == 'param':
                    self.netlist.set_parameter(self.iter_list[iter_no].elem, value)
                elif self.iter_list[iter_no].what == 'component':
                    self.netlist.set_component_value(self.iter_list[iter_no].elem, value)
                elif self.iter_list[iter_no].what == 'model':
                    self.netlist.set_element_model(self.iter_list[iter_no].elem, value)
                else:
                    # TODO: develop other types of sweeps EX: add .STEP instruction
                    raise ValueError("Not Supported sweep")
                iter_no += 1
            if iter_no < 0:
                break

            run_filename = filenamer(**self.current_values) if filenamer else None

            task = self.runner.run(self.netlist, callback=callback, callback_args=callback_args,
                                   switches=switches, timeout=timeout, run_filename=run_filename, exe_log=exe_log)

            # Now storing the simulation information
            if task and task.netlist_file:
                sim_info = self.current_values.copy()
                sim_info['netlist'] = task.netlist_file.name
                self.sim_info[task.runno] = sim_info

            iter_no = len(self.iter_list) - 1  # Resets the counter to start next iteration
        if wait_completion:
            # Now waits for the simulations to end
            self.runner.wait_completion()

    def export_step_info(self, export_filename: Union[Path, str], delimiter: str = ";"):
        """
        Exports the stepping values to a CSV file. It writes a row per each simulation done.
        The columns are all the values that were set during the session. The value on each row is the value
        of the parameter or component value/model at each simulation.
        This information can also be accessed using the sim_info attribute. The sim_info is a dictionary where the
        keys are the simulation number (runno) and the values are an enclosed dictionary containing the parameters
        there were set for that simulation.

        :param export_filename: export file path
        :type export_filename: str or pathlib.Path
        :param delimiter: delimiter character on the CSV
        :type delimiter: str
        """
        import csv

        rows = [runno for runno in self.sim_info]
        rows.sort()

        # Extract column names from the first dictionary
        fieldnames = ['runno'] + list(self.sim_info[rows[0]].keys())

        # Open a CSV file for writing
        with open(export_filename, 'w', newline='') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames, delimiter=delimiter,
                                    quotechar='"', quoting=csv.QUOTE_NONNUMERIC)

            # Write the header
            writer.writeheader()

            # Write the data
            for runno in rows:
                row_data_with_id = {'runno': runno}
                row_data_with_id.update(self.sim_info[runno])
                writer.writerow(row_data_with_id)

    # def run(self, netlist: Union[str, Path, BaseEditor], *,
    #         wait_resource: bool = True,
    #         callback: Union[Type[ProcessCallback], Callable] = None,
    #         callback_args: Union[tuple, dict] = None,
    #         switches=None,
    #         timeout: float = None,
    #         run_filename: str = None,
    #         exe_log: bool = False) -> Union[RunTask, None]:
    #     """Rather uses run_all instead"""
    #     self.run_all()

    @property
    def okSim(self):
        """Number of successful simulations"""
        return self.runner.okSim

    @property
    def runno(self):
        """Number simulations done."""
        return self.runner.runno

    @property
    def failSim(self):
        """Number of failed simulations"""
        return self.runner.failSim



import dataclasses
import enum
from copy import deepcopy
from typing import Union

UpdateValueType = Union[str, int, float, None]


class UpdateType(enum.Enum):
    """The UpdateType holds the information about what is being updated."""
    InvalidUpdate = 0
    UpdateParameter = enum.auto()
    UpdateComponentValue = enum.auto()
    UpdateComponentParameter = enum.auto()
    DeleteParameter = enum.auto()
    DeleteComponent = enum.auto()
    DeleteComponentParameter = enum.auto()
    DeleteInstruction = enum.auto()
    AddParameter = enum.auto()
    AddComponent = enum.auto()
    AddComponentParameter = enum.auto()
    AddInstruction = enum.auto()
    CloneSubcircuit = enum.auto()


@dataclasses.dataclass
class Update:
    """An object containing an update element."""
    name: str
    value: UpdateValueType
    updates: UpdateType = UpdateType.InvalidUpdate

    def __repr__(self):
        if self.updates == UpdateType.UpdateParameter:
            return f"Parameter {self.name} was updated to {self.value}"
        elif self.updates == UpdateType.UpdateComponentValue:
            return f"Component {self.name} value was updated to {self.value}"
        elif self.updates == UpdateType.UpdateComponentParameter:
            return f"Parameter {self.name} was updated to {self.value}"
        elif self.updates == UpdateType.DeleteParameter:
            return f"Parameter {self.name} was deleted"
        elif self.updates == UpdateType.DeleteComponent:
            return f"Component {self.name} was deleted"
        elif self.updates == UpdateType.DeleteComponentParameter:
            return f"Component Parameter {self.name} was deleted"
        elif self.updates == UpdateType.DeleteInstruction:
            return f"Instruction \"{self.value}\" was deleted"
        elif self.updates == UpdateType.AddParameter:
            return f"Parameter {self.name} was added with {self.value}"
        elif self.updates == UpdateType.AddComponent:
            return f"Component {self.name} was added"
        elif self.updates == UpdateType.AddComponentParameter:
            return f"Component Parameter {self.name} was added with value {self.value}"
        elif self.updates == UpdateType.AddInstruction:
            return f"Instruction \"{self.value}\" was added."
        elif self.updates == UpdateType.CloneSubcircuit:
            return f"Sub-circuit {self.name} was added"
        else:
            return "Invalid Update"


class Updates:
    """A list of updates done to a Netlist"""
    def __init__(self):
        self.netlist_updates: list[Update] = []

    def __copy__(self):
        newone = type(self)()
        newone.netlist_updates = deepcopy(self.netlist_updates)
        return newone

    def __len__(self):
        return len(self.netlist_updates)

    def __getitem__(self, item: Union[int, slice, str]) -> Union[Update, list[Update]]:
        if isinstance(item, (int, slice)):
            return self.netlist_updates[item]  # with item = slice, we could get a list here. Otherwise, it is a single Update
        elif isinstance(item, str):
            # Try to get it by name
            for update in self.netlist_updates:
                if update.name == item:
                    return update
            raise IndexError(f"The item {item} doesn't exit.")
        else:
            raise TypeError("getitem only supports int, slice and str")

    def clear(self):
        """Clear the list of updates."""
        self.netlist_updates.clear()

    def add_update(self, name: str, value: UpdateValueType, updates: UpdateType):
        """Add an update to the list"""
        for update in self.netlist_updates:
            if (update.name == name and
                    (name != "INSTRUCTION" or value == update.value) and  # if instruction then it should match
                    (update.updates == updates or updates == UpdateType.InvalidUpdate)):
                break
        else:
            update = Update(name, value, updates)
            self.netlist_updates.append(update)
        if updates != UpdateType.InvalidUpdate:
            update.updates = updates
        update.value = value
        return update

    def value(self, reference) -> UpdateValueType:
        "Get the value update done to a component. Returns None if there wasn't any update."
        for update in self.netlist_updates:
            if update.updates == UpdateType.UpdateComponentValue and update.name == reference:
                return update.value

    def parameter(self, name):
        "Get the update done to a parameter. Returns None if there wasn't any update."
        for update in self.netlist_updates:
            if update.updates in (UpdateType.UpdateParameter, UpdateType.AddParameter) and name == update.name:
                return update.value



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union

import numpy as np

from spicelib.raw.raw_classes import Axis, TraceRead

# Constants
MIN_BYTES_IN_FILE = 20

# base class that has the minimum for a plot. Inherited by the PlotData and RawRead classes
class PlotInterface(ABC):

    @property
    @abstractmethod
    def nVariables(self) -> int:
        """Number of variables in the RAW file
        """
        ...

    @property
    @abstractmethod
    def nPoints(self) -> int:
        """Number of points in the RAW file
        """
        ...

    @property
    @abstractmethod
    def raw_type(self) -> str:
        """The type of the RAW file, either 'binary:' or 'values:'"""
        ...

    @property
    @abstractmethod
    def aliases(self) -> dict[str, str]:
        """QSpice defines aliases for some of the traces that can be computed from other traces.
        """
        ...

    @property
    @abstractmethod
    def backannotations(self) -> list[str]:
        """List to store the backannotations found in the RAW file header
        """
        ...

    @property
    @abstractmethod
    def has_axis(self) -> bool:
        """Indicates if the RAW file has an axis.
        This is True for all RAW file plots except for 'Operating Point', 'Transfer Function', and 'Integrated Noise'.
        """
        ...

    @property
    @abstractmethod
    def flags(self) -> list[str]:
        """List of Flags that are used in this plot. See :doc:`../varia/raw_file` for details.
        """
        ...

    @property
    @abstractmethod
    def steps(self) -> Union[list[dict[str, int]], None]:
        """List of steps in the RAW file, if it exists.
        If the RAW file does not contain stepped data, this will be None.
        If the RAW file contains stepped data, this will be a list of step numbers.
        """
        ...

    @abstractmethod
    def get_raw_property(self, property_name=None) -> Union[str, dict[str, str]]:
        """
        Get a property. By default, it returns all properties defined in the RAW file.

        :param property_name: name of the property to retrieve. If None, all properties are returned.
        :type property_name: str
        :returns: Property object
        :rtype: str
        :raises: ValueError if the property doesn't exist
        """
        ...

    @abstractmethod
    def get_raw_properties(self) -> dict[str, str]:
        """
        Get all raw properties.

        :return: Dictionary of all raw properties
        :rtype: dict[str, str]
        """
        ...

    @abstractmethod
    def get_plot_name(self) -> str:
        """Returns the type of the plot read from the RAW file. See :doc:`../varia/raw_file` for details."""
        ...

    @abstractmethod
    def get_trace_names(self) -> list[str]:
        """Returns a list of trace names in the plot."""
        ...

    @abstractmethod
    def get_trace(self, trace_ref: Union[str, int]) -> Union[Axis, TraceRead]:
        """
        Retrieves the trace with the requested name (trace_ref).

        :param trace_ref: Name of the trace or the index of the trace
        :type trace_ref: str or int
        :return: An object containing the requested trace
        :rtype: DataSet subclass
        :raises IndexError: When a trace is not found
        """
        ...

    @abstractmethod
    def get_wave(self, trace_ref: Union[str, int], step: int = 0) -> np.ndarray:
        """
        Retrieves the trace data with the requested name (trace_ref), optionally providing the step number.

        :param trace_ref: Name of the trace or the index of the trace
        :type trace_ref: str or int
        :param step: Optional parameter specifying which step to retrieve.
        :type step: int
        :return: A numpy array containing the requested waveform.
        :rtype: numpy.array
        :raises IndexError: When a trace is not found
        """
        ...

    @abstractmethod
    def get_axis(self, step: int = 0) -> Union[np.ndarray, list[float]]:
        """
        This function is equivalent to get_trace(0).get_wave(step) instruction.
        It also implements a workaround on a LTSpice issue when using 2nd Order compression, where some values on
        the time trace have a negative value.

        :param step: Step number, defaults to 0
        :type step: int, optional
        :raises RuntimeError: if the RAW file does not have an axis.
        :return: Array with the X axis
        :rtype: Union[np.ndarray, list[float]]
        """
        ...

    @abstractmethod
    def get_len(self, step: int = 0) -> int:
        """
        Returns the length of the data at the give step index.

        :param step: the step index, defaults to 0
        :type step: int, optional
        :return: The number of data points
        :rtype: int
        """
        ...

    @abstractmethod
    def get_steps(self, **kwargs) -> Union[list[int], range]:
        """Returns the steps that correspond to the query set in the `**kwargs` parameters.
        Example: ::

            raw_read.get_steps(V5=1.2, TEMP=25)

        This will return all steps in which the voltage source V5 was set to 1.2V and the TEMP parameter is 24 degrees.
        This feature is only possible if a .log file with the same name as the .raw file exists in the same directory.
        Note: the correspondence between step numbers and .STEP information is stored on the .log file.

        :key kwargs:
         key-value arguments in which the key correspond to a stepped parameter or source name, and the value is the
         stepped value.

        :return: The steps that match the query
        :rtype: list[int]
        """
        ...

    @abstractmethod
    def export(self, columns: Union[list, None] = None, step: Union[int, list[int]] = -1, **kwargs) -> dict[str, list]:
        """
        Returns a native python class structure with the requested trace data and steps.
        It consists of an ordered dictionary where the columns are the keys and the values are lists with the data.

        This function is used by the export functions.

        :param step: Step number to retrieve. If not given, it will return all steps
        :type step: int
        :param columns: List of traces to use as columns. Default is all traces
        :type columns: list
        :param kwargs: Additional arguments to pass to the pandas.DataFrame constructor
        :type kwargs: ``**dict``
        :return: A pandas DataFrame
        :rtype: pandas.DataFrame
        """
        ...

    @abstractmethod
    def to_dataframe(self, columns: Union[list, None] = None, step: Union[int, list[int]] = -1, **kwargs):
        """
        Returns a pandas DataFrame with the requested data.

        :param step: Step number to retrieve. If not given, it
        :type step: int
        :param columns: List of traces to use as columns. Default is all traces
        :type columns: list
        :param kwargs: Additional arguments to pass to the pandas.DataFrame constructor
        :type kwargs: ``**dict``
        :return: A pandas DataFrame
        :rtype: pandas.DataFrame
        """
        # cannot type the return values, as pandas is an optional dependency
        ...

    @abstractmethod
    def to_csv(self, filename: Union[str, Path], columns: Union[list, None] = None, step: Union[int, list[int]] = -1,
               separator=',', **kwargs):
        """
        Saves the data to a CSV file.

        :param filename: Name of the file to save the data to
        :type filename: str
        :param columns: List of traces to use as columns. Default is all traces
        :type columns: list
        :param step: Step number to retrieve. If not given, it
        :type step: int
        :param separator: separator to use in the CSV file
        :type separator: str
        :param kwargs: Additional arguments to pass to the pandas.DataFrame.to_csv function
        :type kwargs: ``**dict``
        """
        ...

    @abstractmethod
    def to_excel(self, filename: Union[str, Path], columns: Union[list, None] = None, step: Union[int, list[int]] = -1,
                 **kwargs):
        """
        Saves the data to an Excel file.

        :param filename: Name of the file to save the data to
        :type filename: Union[str, pathlib.Path]
        :param columns: List of traces to use as columns. Default is None, meaning all traces
        :type columns: list, optional
        :param step: Step number to retrieve, defaults to -1
        :type step: Union[int, list[int]], optional
        :param kwargs: Additional arguments to pass to the pandas.DataFrame.to_excel function
        :type kwargs: ``**dict``
        :raises ImportError: when the 'pandas' module is not installed
        """
        ...


